import itertools
import time
from multiprocessing import Process, Queue

import numpy as np

__author__ = "Can Kizilkale"
__date__ = "12/30/21"

prfp = 5


def Reconstruct(
    df_input,
    Algchoice="FPNA",
    auto_tune=1,
    overlapp_coeff=0.3,
    hist_coeff=20,
    fnfp=100,
    fnc=1,
    postprcoef=5,
    post_fp=0.0001,
    post_fn=0.1,
    n_proc=7,
):
    global prfp
    prfp = postprcoef
    q = Queue()
    matrix_input = ReadFileNA(df_input.copy())
    matrix_input_raw = ReadFasis(df_input.copy())
    matrix_NA_est = Estimated_Matrix(df_input.copy())
    h_range = list(range(1, 100, 2))
    oc_range = [x / 10 for x in range(1, 5)]
    tune_var = itertools.product(h_range, oc_range)

    running_time = 0
    matrix_recons = None

    if Algchoice == "FN":
        s_time = time.time()
        matrix_recons = greedyPtreeNew(matrix_input.astype(bool))[1]
        e_time = time.time()
        running_time = e_time - s_time

    if Algchoice == "FPNA" and auto_tune == 0:
        s_time = time.time()
        apprx_ordr = sum(matrix_NA_est)
        matrix_recons = greedyPtreeNA(
            matrix_input.astype(bool), apprx_ordr, overlapp_coeff, hist_coeff
        )[2]
        e_time = time.time()
        running_time = e_time - s_time

    if Algchoice == "FPNA" and auto_tune == 1:
        tune_var = list(tune_var)
        proc_size = np.ceil(len(tune_var) / n_proc).astype(int)
        cpu_range = []

        for i in range(n_proc):
            s_i = i * proc_size
            e_i = s_i + proc_size
            if e_i > len(tune_var):
                e_i = len(tune_var)
            cpu_range.append(tune_var[s_i:e_i])
        s_time = time.time()
        p = [
            Process(
                target=Auto_fnfp,
                args=(
                    q,
                    cpu_range[i],
                    matrix_input,
                    matrix_NA_est,
                    matrix_input_raw,
                    fnfp,
                    fnc,
                ),
            )
            for _ in range(len(cpu_range))
        ]
        for i in p:
            i.start()
        for i in p:
            i.join()
        ret = []
        while not q.empty():
            ret.append(q.get())

        [m_r, d_min] = ret[0]
        matrix_recons = m_r
        for i in range(1, len(ret)):
            [m, d] = ret[i]
            if d < d_min:
                matrix_recons = m
                d_min = d
        q.empty()
        e_time = time.time()
        running_time = e_time - s_time

    matrix_recons = postprocess_col(df_input, matrix_recons, pfn=post_fn, pfp=post_fp)

    return matrix_recons, running_time


def Auto_fnfp(q, tune_ran, m_input, m_NA_est, m_input_raw, fnfp, fnc):
    apprx_ordr = sum(m_NA_est)

    matrix_recon = greedyPtreeNA(
        m_input.astype(bool), apprx_ordr, tune_ran[0][1], tune_ran[0][0]
    )[2]
    matrix_rec_Temp = deleteNas(m_input_raw, matrix_recon)
    n_10 = np.sum(matrix_rec_Temp < m_input)
    n_01 = np.sum(matrix_rec_Temp > m_input)

    distance = fnfp * n_10 + fnc * n_01

    for [h_i, overlapp_coeff] in tune_ran:
        matrix_rec_i = greedyPtreeNA(
            m_input.astype(bool), apprx_ordr, overlapp_coeff, h_i
        )[2]
        matrix_rec_Temp = deleteNas(m_input_raw, matrix_rec_i)
        n_10 = np.sum(matrix_rec_Temp < m_input, dtype="int64")
        n_01 = np.sum(matrix_rec_Temp > m_input, dtype="int64")
        distance_i = fnfp * n_10 + fnc * n_01
        if distance_i < distance:
            matrix_recon = matrix_rec_i.copy()
            distance = distance_i
    q.full()
    q.put([matrix_recon, distance])


def deleteNas(M_in, M_out):
    M_o = M_out.copy()
    NA_position = np.argwhere(M_in == 3)
    for j in NA_position:
        M_o[j[0], j[1]] = 0
    return M_o


def greedyPtreeNew(M_input):
    M_copy = M_input.copy()

    ISet1 = np.argsort(sum(M_copy))
    ISet = []
    for i in range(M_copy.shape[1]):
        ISet.append(ISet1[i])

    bret = []
    while len(ISet) > 1:
        pivot_index = ISet[-1]
        Sremaining = ISet.copy()
        pivot_vector = M_copy[:, pivot_index]
        cum_vector = np.copy(pivot_vector)
        while_cont = 1

        while while_cont == 1:
            while_cont = 0

            for j in Sremaining:
                cap_j = cum_vector * M_copy[:, j]

                if np.any(cap_j):
                    cum_vector = cum_vector + M_copy[:, j]
                    while_cont = 1
                    Sremaining.remove(j)
        M_copy[:, pivot_index] = cum_vector
        ISet.remove(pivot_index)
        bret = np.argwhere(M_copy.astype(int) > M_input.astype(int))
    return [bret, M_copy]


def greedyPtreeNA(M_input, approx_order, oc, hc):
    overlap_coeff = oc
    hist_coeff = hc

    M_copy = M_input.copy()
    ISet1 = np.argsort(approx_order)
    ISet = []
    for i in range(M_copy.shape[1]):
        ISet.append(ISet1[i])

    bret = []
    pret = []

    while len(ISet) > 1:
        pivot_index = ISet[-1]
        Sremaining = ISet.copy()
        pivot_vector = M_copy[:, pivot_index]
        while_cont = 1
        cum_hist = np.zeros(M_input.shape[0])

        while while_cont == 1:
            while_cont = 0
            for j in Sremaining:
                cap_i = pivot_vector * M_copy[:, j]
                min_vec_size = np.min([np.sum(M_copy[:, j]), np.sum(pivot_vector)])
                cap_size = np.sum(cap_i)

                if cap_size / min_vec_size > overlap_coeff:
                    cum_hist = cum_hist + M_copy[:, j].astype(int)
                    while_cont = 1
                    Sremaining.remove(j)

        cnumT = np.floor(cum_hist.max() / hist_coeff)
        cum_vector = cum_hist > cnumT

        for j in ISet:
            capj = cum_vector * M_copy[:, j]
            difj = M_copy[:, j] > capj
            if np.sum(capj) > np.sum(difj):
                M_copy[:, j] = capj

            else:
                M_copy[:, j] = difj

        M_copy[:, pivot_index] = cum_vector
        ISet.remove(pivot_index)

    bret = np.argwhere(M_copy.astype(int) > M_input.astype(int))
    pret = np.argwhere(np.argwhere(M_copy.astype(int) < M_input.astype(int)))
    return [bret, pret, M_copy]


def ReadFfile(df):
    M = df.values.astype(bool)
    return M


def ReadFileNA(df):
    M = df.values
    NA_position = np.argwhere(M == 3)
    for j in NA_position:
        M[j[0], j[1]] = 0
    return M.astype(bool)


def Estimated_Matrix(df):
    M = df.values.astype(float)
    for i in range(M.shape[1]):
        if np.sum(M[:, i] != 3) == 0:
            one_ratio = 0
        else:
            one_ratio = np.sum(M[:, i] == 1) / np.sum(M[:, i] != 3)
        for j in range(M.shape[0]):
            if M[j, i] == 3:
                M[j, i] = one_ratio
    return M


def ReadFasis(df):
    M = df.values
    return M


def postprocess_col(df_input, matrix_recons, pfn, pfp):
    M_noisy = ReadFasis(df_input.copy())
    M_nds = matrix_recons.copy().astype(bool)
    Mtemp = c_m_col(ReadFasis(df_input.copy()), M_nds, pc_fn=pfn, pc_fp=pfp)
    Mtemp2 = Mtemp.copy()
    d10min = np.sum(Mtemp < (M_noisy == 1))
    imp = 1
    while imp:
        Mtemp2 = c_m_row(ReadFasis(df_input.copy()), Mtemp2, pc_fn=pfn, pc_fp=pfp)
        Mtemp2 = c_m_col(ReadFasis(df_input.copy()), Mtemp2, pc_fn=pfn, pc_fp=pfp)

        d10c = np.sum(Mtemp2 < (M_noisy == 1))
        if d10c < d10min:
            d10min = d10c
            Mtemp = Mtemp2.copy()
        else:
            imp = 0
    return Mtemp


def f_d_col(node_piv, M_samples, p_fp=0.005, p_fn=0.1):
    M_nodes = M_samples.copy()
    D10 = (M_nodes.T == 0).astype(int).dot(node_piv == 1)
    D11 = (M_nodes.T == 1).astype(int).dot(node_piv == 1)
    D00 = (M_nodes.T == 0).astype(int).dot(node_piv == 0)
    D01 = (M_nodes.T == 1).astype(int).dot(node_piv == 0)
    distances = -(
        np.multiply(np.log(p_fp), D10)
        + np.multiply(np.log(1 - p_fp), D00)
        + np.multiply(np.log(p_fn), D01)
        + np.multiply(np.log(1 - p_fn), D11)
    )
    return distances


def f_d_row(node_piv, M_samples, p_fp=0.005, p_fn=0.1):
    M_nodes = M_samples.copy()
    distances = np.zeros(M_nodes.shape[0])
    D10 = (M_nodes == 0).astype(int).dot(node_piv == 1)
    D11 = (M_nodes == 1).astype(int).dot(node_piv == 1)
    D00 = (M_nodes == 0).astype(int).dot(node_piv == 0)
    D01 = (M_nodes == 1).astype(int).dot(node_piv == 0)
    distances = -(
        np.multiply(np.log(p_fp), D10)
        + np.multiply(np.log(1 - p_fp), D00)
        + np.multiply(np.log(p_fn), D01)
        + np.multiply(np.log(1 - p_fn), D11)
    )
    return distances


def c_m_col(M_input, M_nodes, pc_fp=0.0001, pc_fn=0.1):
    M_out = M_input.copy()
    for i in range(M_input.shape[1]):
        pivot_v = M_input[:, i]
        distance_i = f_d_col(pivot_v, M_nodes, p_fn=pc_fn, p_fp=pc_fp)
        min_index = np.argmin(distance_i)
        M_out[:, i] = M_nodes[:, min_index]
    return M_out


def c_m_row(M_input, M_nodes, pc_fp=0.0001, pc_fn=0.1):
    M_out = M_input.copy()
    for i in range(M_input.shape[0]):
        pivot_v = M_input[i, :]
        distance_i = f_d_row(pivot_v, M_nodes, p_fn=pc_fn, p_fp=pc_fp)
        min_index = np.argmin(distance_i)
        M_out[i, :] = M_nodes[min_index, :]
    return M_out



import subprocess

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run BWA.")
@click.argument(
    "indir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "outdir",
    required=True,
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.option(
    "--time",
    default="0-10:00:00",
    type=str,
    show_default=True,
    help="Time.",
)
@click.option(
    "--mem",
    default="90",
    type=str,
    show_default=True,
    help="Memory (mem must be at least `n_threads * 10GB`).",
)
@click.option(
    "--n_threads",
    default="8",
    type=str,
    show_default=True,
    help="Number of threads.",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
@click.option(
    "--is_pdx",
    default=False,
    type=bool,
    show_default=True,
    is_flag=True,
    help="Is the the PDX.",
)
def bwa(indir, outdir, ref, time, mem, n_threads, afterok, is_pdx):
    """Run bwa-mem.

    scphylo caller bwa /path/to/in/dir /path/to/out/dir hg19|hg38|mm10 --is_pdx

    FastQ files (*.fastq.gz) --> BAM files (*.mapped.bam)

    if --is_pdx: FastQ files (*.fastq.gz) --> FastQ files (*.fastq.gz)
    """
    scp.ul.mkdir(outdir)
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    def get_command(sample, is_paired):
        cmds = ""
        cmds += cmd([f"module load {scp.settings.tools['bwa']}"])
        cmds += cmd([f"module load {scp.settings.tools['samtools']}"])
        if is_paired:
            infqs = [f"{indir}/{sample}_{i}.fastq.gz" for i in range(1, 3)]
        else:
            infqs = [f"{indir}/{sample}.fastq.gz"]
        cmds += cmd(
            [
                "bwa mem",
                f"-t {n_threads}",
                "-M",
                f"{config['bwa_ref']}",
                f"{' '.join(infqs)}",
                "|",
                f"samtools sort -@{n_threads} -m 10000000000",
                f"-o {outdir}/{sample}.mapped.bam -",
            ]
        )
        if is_pdx:
            cmds += cmd(
                [
                    "samtools view",
                    "-b -f 4",
                    f"{outdir}/{sample}.mapped.bam",
                    f"> {outdir}/{sample}.unmapped.bam",
                ]
            )
            cmds += cmd(
                [
                    "samtools fastq",
                    f"-1 {outdir}/{sample}_1.fastq",
                    f"-2 {outdir}/{sample}_2.fastq",
                    "-0 /dev/null -s /dev/null -n",
                    f"{outdir}/{sample}.unmapped.bam",
                ]
            )
            cmds += cmd(
                [
                    "gzip",
                    f"{outdir}/{sample}_1.fastq",
                ]
            )
            cmds += cmd(
                [
                    "gzip",
                    f"{outdir}/{sample}_2.fastq",
                ]
            )
            cmds += cmd(
                [
                    "rm -rf",
                    f"{outdir}/{sample}.mapped.bam",
                    f"{outdir}/{sample}.unmapped.bam",
                ]
            )
        cmds += cmd(["echo Done!"], islast=True)
        return cmds

    temp = scp.ul.is_paired(indir)
    df_cmds = pd.DataFrame(temp)
    if df_cmds.shape[0] != df_cmds["sample"].nunique():
        scp.logg.error("Samples are not unique!")
    df_cmds["cmd"] = df_cmds.apply(
        lambda x: get_command(x["sample"], x["is_paired"]), axis=1
    )

    cmdmain = write_cmds_get_main(
        df_cmds,
        "bwa",
        time,
        mem,
        None,
        n_threads,
        scp.settings.tools["email"],
        f"{outdir}/_tmp",
        afterok,
    )
    code = subprocess.getoutput(cmdmain)
    scp.logg.info(code)

    return None



import subprocess

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run fastp.")
@click.argument(
    "indir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "outdir",
    required=True,
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
)
@click.option(
    "--time",
    default="0-02:00:00",
    type=str,
    show_default=True,
    help="Time.",
)
@click.option(
    "--mem",
    default="10",
    type=str,
    show_default=True,
    help="Memory.",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def fastp(indir, outdir, time, mem, afterok):
    """Run fastp.

    scphylo caller fastp /path/to/in/dir /path/to/out/dir

    FastQ files --> FastQ files
    """
    scp.ul.mkdir(outdir)

    def get_command(sample, is_paired):
        cmds = ""
        cmds += cmd([f"module load {scp.settings.tools['fastp']}"])
        if is_paired:
            infqs = [
                f"-i {indir}/{sample}_1.fastq.gz",
                f"-I {indir}/{sample}_2.fastq.gz",
            ]
            outfqs = [
                f"-o {outdir}/{sample}_1.fastq.gz",
                f"-O {outdir}/{sample}_2.fastq.gz",
            ]
        else:
            infqs = [
                f"-i {indir}/{sample}.fastq.gz",
            ]
            outfqs = [
                f"-o {outdir}/{sample}.fastq.gz",
            ]
        cmds += cmd(
            [
                "fastp",
                "-w 8",
                f"{' '.join(infqs)}",
                f"{' '.join(outfqs)}",
                f"-h {outdir}/{sample}.fastp.html",
                f"-j {outdir}/{sample}.fastp.json",
                f"> {outdir}/{sample}.fastp.log",
            ]
        )
        cmds += cmd(
            [
                "rm -rf",
                f"{outdir}/{sample}.fastp.html",
                f"{outdir}/{sample}.fastp.json",
                f"{outdir}/{sample}.fastp.log",
            ]
        )
        cmds += cmd(["echo Done!"], islast=True)
        return cmds

    temp = scp.ul.is_paired(indir)
    df_cmds = pd.DataFrame(temp)
    if df_cmds.shape[0] != df_cmds["sample"].nunique():
        scp.logg.error("Samples are not unique!")
    df_cmds["cmd"] = df_cmds.apply(
        lambda x: get_command(x["sample"], x["is_paired"]), axis=1
    )

    cmdmain = write_cmds_get_main(
        df_cmds,
        "fastp",
        time,
        mem,
        None,
        8,
        scp.settings.tools["email"],
        f"{outdir}/_tmp",
        afterok,
    )
    code = subprocess.getoutput(cmdmain)
    scp.logg.info(code)

    return None



import csv
import os

import scphylo as scp


def write_cmds_get_main(
    df, jobname, time, mem, module, nthread, email, tmpdir, afterok=None
):
    cmdfile = f"{tmpdir}/cmd/{jobname}.sh"
    logdir = f"{tmpdir}/log/{jobname}"
    scp.ul.mkdir(f"{tmpdir}/cmd")
    scp.ul.mkdir(logdir)
    if "biowulf" in os.uname()[1]:
        if df.shape[0] > 1:
            cmd = [
                "swarm",
                f"--file {cmdfile}",
                f"--time {time}",
                f"--gb-per-process {mem}",
                f"--logdir {logdir}",
                "--bundle 1",
                "--partition ccr,norm",
                "--processes-per-subjob 1",
                f"--threads-per-process {nthread}",
                "--noht",
            ]
            if module is not None:
                cmd += [
                    f"--module {module}",
                ]
            tmp = [f'--sbatch "--mail-type=END --mail-user={email}']
            if afterok is not None:
                tmp += [f"--dependency=afterok:{afterok}"]
            if jobname == "sra":
                tmp += ["--gres=lscratch:20"]
            tmp += [f'--job-name={jobname}"']
            cmd += tmp
        else:
            cmd = [
                "sbatch",
                "--ntasks=1",
                f"--cpus-per-task={nthread}",
                "--ntasks-per-core=1",
                f"--time={time}",
                f"--mem={mem}g",
                f"--job-name={jobname}",
                "--partition=ccr,norm",
                f"--chdir={tmpdir}",
                f"--output={logdir}/log.o",
                f"--error={logdir}/log.e",
                "--mail-type=END",
                f"--mail-user={email}",
            ]
            if afterok is not None:
                cmd += [f"--dependency=afterok:{afterok}", f"{cmdfile}"]
            else:
                cmd += [f"{cmdfile}"]
    else:
        cmd = [
            f"cat {cmdfile} | parallel -j0 --bar {{}}",
        ]
    cmdmain = " ".join(cmd)
    with open(cmdfile, "w") as out:
        out.write("#!/bin/bash\n\n#" + cmdmain + "\n\n")
        df["cmd"].to_csv(
            out, index=False, header=False, quoting=csv.QUOTE_NONE, sep="\t"
        )
    return cmdmain


def cmd(arr, islast=False):
    if islast:
        return " ".join(arr)
    else:
        return " ".join(arr) + " && "



import glob
import subprocess

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run bam-readcount.")
@click.argument(
    "outdir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "infile",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.option(
    "--time",
    default=["20:00:00", "01:00:00"],
    multiple=True,
    show_default=True,
    help="Time List [mapping, collecting].",
)
@click.option(
    "--mem",
    default=["20", "20"],
    multiple=True,
    show_default=True,
    help="Memory List [mapping, collecting].",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def bamreadcount(outdir, infile, ref, time, mem, afterok):
    """Run bamreadcount.

    scphylo caller bamreadcount /path/to/bam/dir /path/to/snv/file hg19|hg38|mm10

    The snv file must contains `CHROM,POS,Allele`

    BAM files (*.markdup_bqsr.bam) --> AnnData file (_bamreadcount.h5ad.gz)
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    def step1(afterok):
        def get_command(sample):
            cmds = ""
            cmds += cmd([f"module load {scp.settings.tools['bamreadcount']}"])
            cmds += cmd(
                [
                    "bam-readcount",
                    "-w 0",
                    f"-f {config['ref']}",
                    f"-l {outdir}/_tmp/bamreadcount.input",
                    f"{outdir}/{sample}.markdup_bqsr.bam",
                    f"> {outdir}/{sample}.bamreadcount",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        df_cmds["cmd"] = df_cmds.apply(lambda x: get_command(x["sample"]), axis=1)
        cmdmain = write_cmds_get_main(
            df_cmds,
            "bamreadcount-1of2",
            time[0],
            mem[0],
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step2(afterok):
        def get_command():
            cmds = ""
            cmds += cmd(
                [
                    "python "
                    f"{scp.ul.get_file('scphylo.ul/bamreadcount.py')} "
                    f"{outdir} "
                    f"{infile}"
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        cmds_tmp = pd.DataFrame()
        cmds_tmp["cmd"] = [get_command()]
        cmdmain = write_cmds_get_main(
            cmds_tmp,
            "bamreadcount-2of2",
            time[1],
            mem[1],
            None,
            2,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    files = glob.glob(f"{outdir}/*.markdup_bqsr.bam")
    temp = []
    for file in files:
        file = file.split("/")[-1].replace(".markdup_bqsr.bam", "")
        temp.append({"sample": file})
    df_cmds = pd.DataFrame(temp)

    scp.ul.mkdir(f"{outdir}/_tmp")
    vcf = pd.read_csv(infile)
    vcf[["CHROM", "POS", "POS"]].to_csv(
        f"{outdir}/_tmp/bamreadcount.input", sep="\t", header=None, index=None
    )

    cmd1 = step1(afterok)
    code1 = subprocess.getoutput(cmd1)
    cmd2 = step2(code1)
    code2 = subprocess.getoutput(cmd2)
    scp.logg.info(code2)

    return None



"""Utils Module."""

from collections import OrderedDict

import click

from scphylo.commands.utils._consensus import consensus
from scphylo.commands.utils._partf import partf
from scphylo.commands.utils._score import score
from scphylo.commands.utils._search import search
from scphylo.commands.utils._trees import cf2newick, cf2tree


class NaturalOrderGroup(click.Group):
    """Command group trying to list subcommands in the order they were added.

    Make sure you initialize the `self.commands` with OrderedDict instance.
    With decorator, use::
        @click.group(cls=NaturalOrderGroup, commands=OrderedDict())
    """

    def list_commands(self, ctx):
        """List command names as they are in commands dict.

        If the dict is OrderedDict, it will preserve the order commands
        were added.
        """
        return self.commands.keys()


@click.group(
    name="utils",
    cls=NaturalOrderGroup,
    commands=OrderedDict(),
    context_settings={"max_content_width": 300, "terminal_width": 300},
)
def cli_utils():
    """Utils module."""
    return None


cli_utils.add_command(score)
cli_utils.add_command(search)
cli_utils.add_command(cf2newick)
cli_utils.add_command(cf2tree)
cli_utils.add_command(consensus)
cli_utils.add_command(partf)



import glob
import os

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run mpileup.")
@click.argument(
    "outdir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.option(
    "--time",
    default="5-00:00:00",
    type=str,
    show_default=True,
    help="Time.",
)
@click.option(
    "--mem",
    default="150",
    type=str,
    show_default=True,
    help="Memory.",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def mpileup(outdir, ref, time, mem, afterok):
    """Run mpileup.

    scphylo caller mpileup /path/to/in/dir hg19|hg38|mm10

    BAM files (*.markdup_bqsr.bam) --> VCF file (_pileupcalls.vcf)
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    elif ref == "m10x":
        config = scp.settings.m10x
    else:
        config = None

    def cmds(samples):
        cmds = ""
        cmds += cmd([f"module load {scp.settings.tools['bcftools']}"])
        files = " ".join([f"{outdir}/{s}.markdup_bqsr.bam" for s in samples])
        cmds += cmd(
            [
                "bcftools mpileup",
                "--min-BQ 30",
                "--min-MQ 0",
                "--count-orphans",
                "--ignore-overlaps",
                "--output-type u",
                "--max-depth 5000",
                '--annotate "DP,AD"',
                "--threads $SLURM_CPUS_PER_TASK",
                f"-f {config['ref']}",
                files,
                "|",
                "bcftools call",
                "--output-type v",
                # "---multiallelic-caller",
                "--consensus-caller",
                "--variants-only",
                f"-o {outdir}/_pileupcalls.vcf",
            ]
        )
        cmds += cmd(["echo Done!"], islast=True)
        return cmds

    files = glob.glob(f"{outdir}/*.markdup_bqsr.bam")
    temp = []
    for file in files:
        file = file.split("/")[-1].replace(".markdup_bqsr.bam", "")
        temp.append(file)
    df_cmds = pd.DataFrame()
    df_cmds["cmd"] = [cmds(temp)]

    cmdmain = write_cmds_get_main(
        df_cmds,
        "mpileup",
        time,
        mem,
        None,
        16,
        scp.settings.tools["email"],
        f"{outdir}/_tmp",
        afterok,
    )
    os.system(cmdmain)

    return None



from libc.stdlib cimport free, malloc


cdef extern from "main.h":
	cdef int main_in_c(int argc, char **argv)

def run_scprob(cmd):
	cdef char **c_argv
	args = [bytes(x, encoding="utf8") for x in cmd]
	c_argv = <char**>malloc(sizeof(char*) * len(args))
	for idx, s in enumerate(args):
		c_argv[idx] = s
	main_in_c(len(args), c_argv)
	return None



import glob
import subprocess

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run PyClone.")
@click.argument(
    "outdir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.option(
    "--time",
    default="0-02:00:00",
    type=str,
    show_default=True,
    help="Time.",
)
@click.option(
    "--mem",
    default="90",
    type=str,
    show_default=True,
    help="Memory.",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def pyclone(outdir, time, mem, afterok):
    """Run PyClone.

    scphylo caller pyclone /path/to/sequenza/folders

    Sequenza folders (*.sequenza) --> PyClone files
    """

    def get_command(sample):
        cmds = ""
        cmds += cmd(["unset PATH"])
        cmds += cmd(["unset PYTHONPATH"])
        cmds += cmd([f"module load {scp.settings.tools['pyclone']}"])
        cmds += cmd(["module load python/2.7"])
        with open(f"{outdir}/{sample}.sequenza/{sample}_cellularity.txt") as file:
            tumor_content = file.readline().strip()
        cmds += cmd(
            [
                "PyClone setup_analysis",
                f"--in_files {outdir}/{sample}.sequenza/{sample}_pyclone.tsv",
                f"--tumour_contents {tumor_content}",
                "--prior major_copy_number",
                f"--working_dir {outdir}/{sample}.pyclone",
            ]
        )
        cmds += cmd(
            [
                "PyClone run_analysis",
                f"--config_file {outdir}/{sample}.pyclone/config.yaml",
            ]
        )
        cmds += cmd(
            [
                "PyClone build_table",
                f"--config_file {outdir}/{sample}.pyclone/config.yaml",
                f"--out_file {outdir}/{sample}.pyclone/loci.tsv",
                "--table_type loci",
            ]
        )
        cmds += cmd(
            [
                "PyClone plot_clusters",
                f"--config_file {outdir}/{sample}.pyclone/config.yaml",
                f"--plot_file {outdir}/{sample}.pyclone/coordinates_cluster_plot.pdf",
                "--plot_type parallel_coordinates",
            ]
        )
        cmds += cmd(
            [
                "PyClone plot_clusters",
                f"--config_file {outdir}/{sample}.pyclone/config.yaml",
                f"--plot_file {outdir}/{sample}.pyclone/density_cluster_plot.pdf",
                "--plot_type density",
            ]
        )
        cmds += cmd(["echo Done!"], islast=True)
        return cmds

    files = glob.glob(f"{outdir}/*.sequenza")
    temp = []
    for file in files:
        file = file.split("/")[-1].replace(".sequenza", "")
        temp.append({"sample": file})
    df_cmds = pd.DataFrame(temp)
    df_cmds["cmd"] = df_cmds.apply(lambda x: get_command(x["sample"]), axis=1)

    cmdmain = write_cmds_get_main(
        df_cmds,
        "pyclone",
        time,
        mem,
        None,
        1,
        scp.settings.tools["email"],
        f"{outdir}/_tmp",
        afterok,
    )
    code = subprocess.getoutput(cmdmain)
    scp.logg.info(code)

    return None



import networkx as nx
import numpy as np

import scphylo as scp


def collapse(tree):
    tc2 = tree.copy()
    root = scp.ul.root_id(tree)
    nnodes = 1
    for _ in range(len(tree.nodes)):
        d_in = tc2.in_degree(tc2)
        d_out = tc2.out_degree(tc2)
        for node in tc2.nodes():
            if d_out[node] == 1 and d_in[node] == 1:
                parent = list(x for x in tc2.predecessors(node))[0]
                child = list(x for x in tc2.successors(node))[0]
                if d_out[parent] < 2 and d_in[parent] == 1:
                    new_node = root + nnodes
                    nnodes += 1
                    new_label = (
                        f"{tc2.nodes[parent]['label']}"
                        f"{tree.graph['splitter_cell']}"
                        f"{tc2.nodes[node]['label']}"
                    )
                    new_edge = (
                        f"{tc2[parent][node]['label']}"
                        f"{tree.graph['splitter_mut']}"
                        f"{tc2[node][child]['label']}"
                    )

                    tc2 = nx.contracted_nodes(tc2, parent, node, self_loops=False)
                    mapping = {parent: new_node}
                    tc2 = nx.relabel_nodes(tc2, mapping)
                    tc2[new_node][child]["label"] = new_edge
                    tc2.nodes[new_node]["label"] = new_label
                    break
    d_in = tc2.in_degree(tc2)
    d_out = tc2.out_degree(tc2)
    nodes = []
    for node in tc2.nodes():
        if d_out[node] == 0:
            nodes.append(node)
    for node in nodes:
        parent = list(x for x in tc2.predecessors(node))[0]
        if d_out[parent] == 1 and d_in[parent] == 1:
            grandparent = list(x for x in tc2.predecessors(parent))[0]

            new_node = root + nnodes
            nnodes += 1
            new_label = (
                f"{tc2.nodes[parent]['label']}"
                f"{tree.graph['splitter_cell']}"
                f"{tc2.nodes[node]['label']}"
            )
            new_edge = (
                f"{tc2[grandparent][parent]['label']}"
                f"{tree.graph['splitter_mut']}"
                f"{tc2[parent][node]['label']}"
            )

            tc2 = nx.contracted_nodes(tc2, parent, node, self_loops=False)
            mapping = {parent: new_node}
            tc2 = nx.relabel_nodes(tc2, mapping)
            tc2[grandparent][new_node]["label"] = new_edge
            tc2.nodes[new_node]["label"] = new_label
    return tc2


def sample_from_tree(tree, ratio, axis="cell"):
    sampled = []
    if axis == "cell":
        for n in tree.nodes:
            if tree.in_degree(n) == 1 and "––" not in tree.nodes[n]["label"]:
                array = tree.nodes[n]["label"].split(tree.graph["splitter_cell"])
                array = np.random.choice(
                    array, replace=False, size=int(ratio * len(array))
                )
                sampled += list(array)
        return sampled
    elif axis == "mut":
        for _, _, lable in tree.edges(data=True):
            array = lable["label"].split(tree.graph["splitter_mut"])
            array = np.random.choice(array, replace=False, size=int(ratio * len(array)))
            sampled += list(array)
        return sampled
    else:
        return None



import os

import click

import scphylo as scp


@click.command(short_help="Run PhISCS-BnB.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.option(
    "--bounding",
    "-b",
    default="real",
    type=click.Choice(["real", "simulated"]),
    show_default=True,
    help="Number of threads.",
)
@click.option(
    "--time_limit",
    "-t",
    default=None,
    type=int,
    show_default=True,
    help="Time limit.",
)
def bnb(genotype_file, bounding, time_limit):
    """PhISCS-BnB.

    A fast branch and bound algorithm for
    the perfect tumor phylogeny reconstruction problem :cite:`PhISCS-BnB`.

    scphylo solver bnb input.SC -b real
    """
    outfile = os.path.splitext(genotype_file)[0]

    scp.settings.verbosity = "info"
    scp.settings.logfile = f"{outfile}.bnb.log"

    df_in = scp.io.read(genotype_file)
    df_out = scp.tl.bnb(df_in, bounding=bounding, time_limit=time_limit)
    scp.io.write(df_out, f"{outfile}.bnb.CFMatrix")

    return None



import anndata as ad
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns


def _colors(alist):
    return mcolors.LinearSegmentedColormap.from_list("", alist)


def heatmap(
    adata, color_attrs=None, layer="X", figsize=(12, 7), vmin=0, vmax=2, rvb=None
):
    """Plot HeatMap.

    For flips: rvb=["#FFFFFF", "#D92347", "#A9D0F5"]
    For CNV: rvb="RdBu_r", vmin=-0.5, vmax=0.5

    Parameters
    ----------
    adata : [type]
        [description]
    color_attrs : [type], optional
        [description], by default None
    layer : str, optional
        [description], by default "X"
    figsize : tuple, optional
        [description], by default (12, 7)
    vmin : [type], optional
        [description], by default None
    vmax : [type], optional
        [description], by default None
    rvb : [type], optional
        [description], by default None
    """
    if color_attrs is not None:
        row_colors = []
        if isinstance(color_attrs, list):
            for attr in color_attrs:
                row_colors.append(adata.obs[attr])
        else:
            row_colors.append(adata.obs[color_attrs])
    else:
        row_colors = None

    if isinstance(adata, ad.AnnData):
        if layer == "X":
            rvb = _colors(["#A9D0F5", "#000000", "#FFFFFF"])
            adatac = adata[:, adata.var.sort_values(["CHROM", "POS"]).index].copy()
            df = adatac.to_df().copy()
            df[df == 3] = 2
            df.index.name = "cells"
            df.columns.name = "mutations"
            chromosoms = {}
            for i in list(range(1, 23, 2)) + ["Y"]:
                chromosoms[f"chr{i}"] = "#969696"
            for i in list(range(2, 23, 2)) + ["X"]:
                chromosoms[f"chr{i}"] = "#252525"
            adatac.var["chrom_color"] = adata.var["CHROM"].map(chromosoms)
            adatac.var["chrom_color"].name = ""
            column_colors = adatac.var["chrom_color"]
        else:
            df = adata.obsm[layer].copy()
            if isinstance(rvb, list):
                rvb = _colors(rvb)
            column_colors = None
    else:
        df = adata
        if isinstance(rvb, list):
            rvb = _colors(rvb)
        column_colors = None

    sns.clustermap(
        df,
        vmin=vmin,
        vmax=vmax,
        metric="euclidean",
        cmap=rvb,
        row_cluster=False,
        col_cluster=False,
        row_colors=row_colors,
        col_colors=column_colors,
        cbar_pos=None,
        figsize=figsize,
        xticklabels=False,
        yticklabels=False,
        colors_ratio=(0.02, 0.02),
        dendrogram_ratio=0,
    )
    # plt.savefig(filepath, bbox_inches="tight", pad_inches=0)


def plot_dist(adata, attr):
    x = adata.obsp[attr][np.triu_indices(adata.shape[0], 1)]
    plt.hist(x, bins=50)



import glob

import pandas as pd

import scphylo as scp


def is_paired(indir):
    files = glob.glob(f"{indir}/*.fastq.gz")
    last_char_set = {x[-len(".fastq.gz") - 2 : -len(".fastq.gz")] for x in files}
    temp = []
    if last_char_set == {"_1", "_2"}:
        for file in files:
            file = file.split("/")[-1].replace(".fastq.gz", "")
            if file[-2:] == "_1":
                temp.append({"sample": file[:-2], "is_paired": True})
    else:
        for file in files:
            file = file.split("/")[-1].replace(".fastq.gz", "")
            temp.append({"sample": file, "is_paired": False})
    return temp


def get_samples_df(outdir, normal=None):
    files = glob.glob(f"{outdir}/*.markdup_bqsr.bam")
    temp = []
    for file in files:
        file = file.split("/")[-1].replace(".markdup_bqsr.bam", "")
        if normal is not None:
            if file != normal:
                temp.append({"sample": file})
        else:
            temp.append({"sample": file})
    df_cmds_tmp = pd.DataFrame(temp)
    if normal is not None:
        scp.logg.info(
            f"Tumor samples: {','.join(df_cmds_tmp['sample'].values)} & Normal sample:"
            f" {normal}"
        )
    return df_cmds_tmp



import glob
import subprocess

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run HaplotypeCaller.")
@click.argument(
    "outdir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.argument(
    "dna_or_rna",
    required=True,
    type=click.Choice(["dna", "rna"]),
)
@click.option(
    "--time",
    default=["20:00:00", "2-00:00:00"],
    multiple=True,
    show_default=True,
    help="Time List [indexing, mapping].",
)
@click.option(
    "--mem",
    default=["90", "90"],
    multiple=True,
    show_default=True,
    help="Memory List [indexing, mapping].",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def hapcaller(outdir, ref, dna_or_rna, time, mem, afterok):
    """Run HaplotypeCaller.

    scphylo caller hapcaller /path/to/in/dir hg19|hg38|mm10 dna|rna

    BAM files (*.markdup_bqsr.bam) --> VCF file (_hapcaller/jointcalls.g.vcf)
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    if ref == "hg19" or ref == "hg38":
        chroms = [f"chr{i}" for i in range(1, 23)] + ["chrX", "chrY", "chrM"]
    elif ref == "mm10":
        chroms = [f"chr{i}" for i in range(1, 20)] + ["chrX", "chrY", "chrM"]

    def step1(afterok):
        def get_command(sample):
            cmds = ""
            cmds += cmd([f"module load {scp.settings.tools['gatk']}"])
            if dna_or_rna == "rna":
                cmds += cmd(
                    [
                        f'gatk --java-options "-Xmx{int(mem[0])-10}g"',
                        "HaplotypeCaller",
                        f"--reference {config['ref']}",
                        f"--input {outdir}/{sample}.markdup_bqsr.bam",
                        f"--output {outdir}/{sample}.hapcaller.g.vcf",
                        "--dont-use-soft-clipped-bases true",
                        "--standard-min-confidence-threshold-for-calling 20",
                        "--emit-ref-confidence GVCF",
                        "--disable-read-filter MappingQualityReadFilter",
                        "--disable-read-filter GoodCigarReadFilter",
                        "--disable-read-filter NotSecondaryAlignmentReadFilter",
                        "--disable-read-filter MappedReadFilter",
                        "--disable-read-filter MappingQualityAvailableReadFilter",
                        "--disable-read-filter",
                        "NonZeroReferenceLengthAlignmentReadFilter",
                        "--disable-read-filter NotDuplicateReadFilter",
                        "--disable-read-filter PassesVendorQualityCheckReadFilter",
                        "--disable-read-filter WellformedReadFilter",
                    ]
                )
            else:
                cmds += cmd(
                    [
                        f'gatk --java-options "-Xmx{int(mem[0])-10}g"',
                        "HaplotypeCaller",
                        f"--reference {config['ref']}",
                        f"--input {outdir}/{sample}.markdup_bqsr.bam",
                        f"--output {outdir}/{sample}.hapcaller.g.vcf",
                        "--dont-use-soft-clipped-bases true",
                        "--standard-min-confidence-threshold-for-calling 20",
                        "--emit-ref-confidence GVCF",
                    ]
                )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        df_cmds["cmd"] = df_cmds.apply(lambda x: get_command(x["sample"]), axis=1)
        cmdmain = write_cmds_get_main(
            df_cmds,
            "hapcaller-1of4",
            time[0],
            mem[0],
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step2(afterok):
        def get_command(chrom):
            cmds = ""
            files = " --variant ".join(
                [f"{outdir}/{s}.hapcaller.g.vcf" for s in df_cmds["sample"].values]
            )
            cmds += cmd([f"module load {scp.settings.tools['gatk']}"])
            cmds += cmd([f"mkdir -p {outdir}/_hapcaller"])
            cmds += cmd(
                [
                    f'gatk --java-options "-Xmx{int(mem[1])-10}g"',
                    "CombineGVCFs",
                    f"--reference {config['ref']}",
                    f"--variant {files}",
                    f"--output {outdir}/_hapcaller/combinedcalls.{chrom}.g.vcf",
                    f"--intervals {chrom}",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        df_cmds_tmp = pd.DataFrame()
        df_cmds_tmp["chrom"] = chroms
        df_cmds_tmp["cmd"] = df_cmds_tmp.apply(
            lambda x: get_command(x["chrom"]), axis=1
        )
        cmdmain = write_cmds_get_main(
            df_cmds_tmp,
            "hapcaller-2of4",
            time[1],
            mem[1],
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step3(afterok):
        def get_command(chrom):
            cmds = ""
            cmds += cmd([f"module load {scp.settings.tools['gatk']}"])
            cmds += cmd([f"mkdir -p {outdir}/_hapcaller"])
            cmds += cmd(
                [
                    f'gatk --java-options "-Xmx{int(mem[1])-10}g"',
                    "GenotypeGVCFs",
                    f"--reference {config['ref']}",
                    f"--variant {outdir}/_hapcaller/combinedcalls.{chrom}.g.vcf",
                    f"--output {outdir}/_hapcaller/jointcalls.{chrom}.g.vcf",
                    f"--intervals {chrom}",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        df_cmds_tmp = pd.DataFrame()
        df_cmds_tmp["chrom"] = chroms
        df_cmds_tmp["cmd"] = df_cmds_tmp.apply(
            lambda x: get_command(x["chrom"]), axis=1
        )
        cmdmain = write_cmds_get_main(
            df_cmds_tmp,
            "hapcaller-3of4",
            time[1],
            mem[1],
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step4(afterok):
        def get_command():
            cmds = ""
            cmds += cmd([f"module load {scp.settings.tools['bcftools']}"])
            files = " ".join(
                [f"{outdir}/_hapcaller/jointcalls.{chrom}.g.vcf" for chrom in chroms]
            )
            cmds += cmd(
                [
                    "bcftools concat",
                    f"-o {outdir}/_hapcaller/jointcalls.g.vcf",
                    f"{files}",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        cmds_tmp = pd.DataFrame()
        cmds_tmp["cmd"] = [get_command()]
        cmdmain = write_cmds_get_main(
            cmds_tmp,
            "hapcaller-4of4",
            time[0],
            mem[0],
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    files = glob.glob(f"{outdir}/*.markdup_bqsr.bam")
    temp = []
    for file in files:
        file = file.split("/")[-1].replace(".markdup_bqsr.bam", "")
        temp.append({"sample": file})
    df_cmds = pd.DataFrame(temp)

    cmd1 = step1(afterok)
    code1 = subprocess.getoutput(cmd1)
    cmd2 = step2(code1)
    code2 = subprocess.getoutput(cmd2)
    cmd3 = step3(code2)
    code3 = subprocess.getoutput(cmd3)
    cmd4 = step4(code3)
    code4 = subprocess.getoutput(cmd4)
    scp.logg.info(code4)

    return None



cdef extern from "mltd.h":
	ctypedef struct MLTDResult:
		int distance
		int similarity
		double normalized_similarity
	cdef MLTDResult calc_mltd(const char* tree1, const char* tree2)

def run_mltd(file1, file2):
	cdef bytes file1_bytes = file1.encode()
	cdef char* cfile1 = file1_bytes
	cdef bytes file2_bytes = file2.encode()
	cdef char* cfile2 = file2_bytes
	cdef MLTDResult result = calc_mltd(cfile1, cfile2)
	result = {
		'distance': result.distance,
		'similarity': result.similarity,
		'normalized_similarity': result.normalized_similarity,
	}
	return result



import click

import scphylo as scp


@click.command(short_help="Calculate scores between two trees.")
@click.argument(
    "ground_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "inferred_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
def score(ground_file, inferred_file):
    """Scores of Anscenstor-descendant, Different-lineage, MLTD and etc.

    scphylo utils score ground.CFMatrix inferred.CFMatrix
    """
    scp.settings.verbosity = "info"

    df_g = scp.io.read(ground_file)
    df_s = scp.io.read(inferred_file)

    gs = scp.tl.gs(df_g, df_s)
    scp.logg.info(f"GS={gs:0.4f}")
    ad = scp.tl.ad(df_g, df_s)
    scp.logg.info(f"AD={ad:0.4f}")
    dl = scp.tl.dl(df_g, df_s)
    scp.logg.info(f"DL={dl:0.4f}")
    mltd = scp.tl.mltd(df_g, df_s)["normalized_similarity"]
    scp.logg.info(f"MLTSM={mltd:0.4f}")
    tpted = scp.tl.tpted(df_g, df_s)
    scp.logg.info(f"TPTED={tpted:0.4f}")
    rf = scp.tl.rf(df_g, df_s)
    scp.logg.info(f"RF={rf:0.4f}")
    caset = scp.tl.caset(df_g, df_s)
    scp.logg.info(f"CASet={caset:0.4f}")
    disc = scp.tl.disc(df_g, df_s)
    scp.logg.info(f"DISC={disc:0.4f}")
    mp3 = scp.tl.mp3(df_g, df_s)
    scp.logg.info(f"MP3={mp3:0.4f}")

    return None



import math
import time
from itertools import combinations

import numpy as np
import pandas as pd
from pysat.examples.rc2 import RC2
from pysat.formula import WCNF
from scipy.stats import binom

import scphylo as scp
from scphylo.external._betabinom import pmf_BetaBinomial


def phiscsb(df_input, alpha, beta, experiment=False):
    """Solving using PhISCS-B (only in single-cell mode, no bulk).

    a combinatorial approach for subperfect tumor phylogeny reconstruction
    via integrative use of single-cell and bulk sequencing data :cite:`PhISCS`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing
        entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    experiment : :obj:`bool`, optional
        Is in the experiment mode (the log won't be shown), by default False

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    if not experiment:
        scp.logg.info(f"running PhISCS-B with alpha={alpha}, beta={beta}")
    cells = list(df_input.index)
    snvs = list(df_input.columns)
    df_input = df_input.replace("?", 3)
    df_input = df_input.astype(int)
    I_mtr = df_input.values

    rc2 = RC2(WCNF())

    num_cells = len(cells)
    num_mutations = len(snvs)

    Y = np.empty((num_cells, num_mutations), dtype=np.int64)
    numVarY = 0
    for i in range(num_cells):
        for j in range(num_mutations):
            numVarY += 1
            Y[i, j] = numVarY

    B = np.empty((num_mutations, num_mutations, 2, 2), dtype=np.int64)
    numVarB = 0
    for p in range(num_mutations):
        for q in range(p + 1, num_mutations):
            for i, j in [(0, 1), (1, 0), (1, 1)]:
                numVarB += 1
                B[p, q, i, j] = numVarY + numVarB

    Z = np.empty((num_cells, num_mutations), dtype=np.int64)
    numVarZ = 0
    for i in range(num_cells):
        for j in range(num_mutations):
            if I_mtr[i, j] == 0:
                numVarZ += 1
                Z[i, j] = numVarY + numVarB + numVarZ

    for p in range(num_mutations):
        for q in range(p + 1, num_mutations):
            rc2.add_clause([-B[p, q, 0, 1], -B[p, q, 1, 0], -B[p, q, 1, 1]])
            for i in range(num_cells):
                rc2.add_clause([-Y[i, p], -Y[i, q], B[p, q, 1, 1]])
                rc2.add_clause([Y[i, p], -Y[i, q], B[p, q, 0, 1]])
                rc2.add_clause([-Y[i, p], Y[i, q], B[p, q, 1, 0]])

    for i in range(num_cells):
        for j in range(num_mutations):
            # 0->1
            if alpha == 0:
                if I_mtr[i, j] == 0:
                    rc2.add_clause([-Y[i, j]], weight=1)
                if I_mtr[i, j] == 1:
                    rc2.add_clause([Y[i, j]])

            # 0->1 and 1->0
            if alpha > 0:
                if I_mtr[i, j] == 0:
                    rc2.add_clause([Y[i, j], Z[i, j]])
                    rc2.add_clause([-Y[i, j], -Z[i, j]])
                    rc2.add_clause([Z[i, j]], weight=math.log((1 - alpha) / beta))
                if I_mtr[i, j] == 1:
                    rc2.add_clause([Y[i, j]], weight=math.log((1 - beta) / alpha))

    s_time = time.time()
    variables = rc2.compute()
    e_time = time.time()
    running_time = e_time - s_time

    sol_Y = np.empty((num_cells, num_mutations), dtype=np.int8)
    numVar = 0
    for i in range(num_cells):
        for j in range(num_mutations):
            sol_Y[i, j] = variables[numVar] > 0
            numVar += 1

    df_output = pd.DataFrame(sol_Y)
    df_output.columns = snvs
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"

    if not experiment:
        scp.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output


def phiscsi(df_input, alpha, beta, time_limit=86400, n_threads=1):
    """Solving using PhISCS-I (only in single-cell mode, no bulk).

    a combinatorial approach for subperfect tumor phylogeny reconstruction
    via integrative use of single-cell and bulk sequencing data :cite:`PhISCS`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing
        entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    time_limit : :obj:`int`, optional
        Time limit of the Gurobi running in seconds, by default 86400 (one day)
    n_threads : :obj:`int`, optional
        Number of threads for Gurobi solver, by default 1

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    gp, gp_is_not_imported = scp.ul.import_gurobi()
    if gp_is_not_imported:
        scp.logg.error("Unable to import a package!")

    scp.logg.info(
        f"running PhISCS-I with alpha={alpha}, beta={beta}, time_limit={time_limit}, "
        f"n_threads={n_threads}"
    )
    cells = list(df_input.index)
    snvs = list(df_input.columns)
    df_input = df_input.replace("?", 3)
    df_input = df_input.astype(int)
    I_mtr = df_input.values

    model = gp.Model("ILP")
    model.Params.OutputFlag = 0
    model.Params.LogFile = ""
    model.Params.Threads = n_threads
    model.Params.TimeLimit = time_limit

    num_cells = len(cells)
    num_mutations = len(snvs)
    Y = {}
    B = {}
    for c in range(num_cells):
        for m in range(num_mutations):
            if alpha == 0:
                # 0->1
                if I_mtr[c, m] == 0:
                    Y[c, m] = model.addVar(
                        vtype=gp.GRB.BINARY, obj=1, name=f"Y({c},{m})"
                    )
                elif I_mtr[c, m] == 1:
                    Y[c, m] = 1
                else:
                    Y[c, m] = model.addVar(
                        vtype=gp.GRB.BINARY, obj=0, name=f"Y({c},{m})"
                    )
            else:
                # 0->1 & 1->0
                Y[c, m] = model.addVar(vtype=gp.GRB.BINARY, name=f"Y({c},{m})")
    for p in range(num_mutations):
        for q in range(p + 1, num_mutations):
            B[p, q, 1, 1] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},1,1]"
            )
            B[p, q, 1, 0] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},1,0]"
            )
            B[p, q, 0, 1] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},0,1]"
            )
    for p in range(num_mutations):
        for q in range(p + 1, num_mutations):
            model.addConstr(B[p, q, 0, 1] + B[p, q, 1, 0] + B[p, q, 1, 1] <= 2)
            for i in range(num_cells):
                model.addConstr(Y[i, p] + Y[i, q] - B[p, q, 1, 1] <= 1)
                model.addConstr(-Y[i, p] + Y[i, q] - B[p, q, 0, 1] <= 0)
                model.addConstr(Y[i, p] - Y[i, q] - B[p, q, 1, 0] <= 0)

    if alpha == 0:
        # 0->1
        model.Params.ModelSense = gp.GRB.MINIMIZE
    else:
        # 0->1 & 1->0
        objective = 0
        for j in range(num_mutations):
            for i in range(num_cells):
                if I_mtr[i, j] == 0:
                    objective += (
                        np.log(1 - alpha) + np.log(beta / (1 - alpha)) * Y[i, j]
                    )
                if I_mtr[i, j] == 1:
                    objective += np.log(alpha) + np.log((1 - beta) / alpha) * Y[i, j]
        model.setObjective(objective, gp.GRB.MAXIMIZE)

    s_time = time.time()
    model.optimize()
    e_time = time.time()
    running_time = e_time - s_time

    sol_Y = np.zeros((num_cells, num_mutations), dtype=np.int8)
    for i in range(num_cells):
        for j in range(num_mutations):
            if alpha == 0:
                if I_mtr[i, j] != 1:
                    sol_Y[i, j] = Y[i, j].X > 0.5
                else:
                    sol_Y[i, j] = 1
            else:
                sol_Y[i, j] = Y[i, j].X > 0.5

    df_output = pd.DataFrame(sol_Y)
    df_output.columns = snvs
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"

    scp.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output


def phiscsb_bulk(
    df_input,
    alpha,
    beta,
    kmax=0,
    vaf_info=None,
    delta=0.2,
):
    """Solving using PhISCS-B (in single-cell mode with bulk and mutation elimination).

    a combinatorial approach for subperfect tumor phylogeny reconstruction
    via integrative use of single-cell and bulk sequencing data :cite:`PhISCS`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing
        entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    kmax : :obj:`int`, optional
        Max number of mutations to be eliminated, by default 0
    vaf_info : :class:`pandas.DataFrame`, optional
        Information about the variant allele frequency in bulk data
        The size is n_SNVs x n_samples, by default None
    delta : :obj:`float`, optional
        Delta parameter accounting for VAF variance, by default 0.2

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    # TODO: implement

    scp.logg.info(
        f"running PhISCS-B-orig with alpha={alpha}, beta={beta}, kmax={kmax}, "
        f"vaf_info={vaf_info}, delta={delta}"
    )

    cells = list(df_input.index)
    snvs = list(df_input.columns)
    df_input = df_input.replace("?", 3)
    df_input = df_input.astype(int)
    I_mtr = df_input.values

    rc2 = RC2(WCNF())

    num_cells = len(cells)
    num_mutations = len(snvs)

    Y = np.empty((num_cells, num_mutations), dtype=np.int64)
    numVarY = 0
    for i in range(num_cells):
        for j in range(num_mutations):
            numVarY += 1
            Y[i, j] = numVarY

    B = np.empty((num_mutations, num_mutations, 2, 2), dtype=np.int64)
    numVarB = 0
    for p in range(num_mutations):
        for q in range(p + 1, num_mutations):
            for i, j in [(0, 1), (1, 0), (1, 1)]:
                numVarB += 1
                B[p, q, i, j] = numVarY + numVarB

    Z = np.empty((num_cells, num_mutations), dtype=np.int64)
    numVarZ = 0
    for i in range(num_cells):
        for j in range(num_mutations):
            if I_mtr[i, j] == 0:
                numVarZ += 1
                Z[i, j] = numVarY + numVarB + numVarZ

    numVarK = 0
    if kmax > 0:
        K = np.empty(num_mutations + 1, dtype=np.int64)
        for j in range(num_mutations + 1):
            numVarK += 1
            K[j] = numVarY + numVarB + numVarZ + numVarK

    if vaf_info is not None:
        A = np.empty((num_cells + 1, num_mutations + 1), dtype=np.int64)
        numVarA = 0
        for i in range(num_cells + 1):
            for j in range(num_mutations + 1):
                numVarA += 1
                A[i, j] = numVarY + numVarB + numVarZ + numVarK + numVarA

    for p in range(num_mutations):
        for q in range(p + 1, num_mutations):
            if kmax > 0:
                rc2.add_clause(
                    [K[p], K[q], -B[p, q, 0, 1], -B[p, q, 1, 0], -B[p, q, 1, 1]]
                )
            else:
                rc2.add_clause([-B[p, q, 0, 1], -B[p, q, 1, 0], -B[p, q, 1, 1]])
            for i in range(num_cells):
                rc2.add_clause([-Y[i, p], -Y[i, q], B[p, q, 1, 1]])
                rc2.add_clause([Y[i, p], -Y[i, q], B[p, q, 0, 1]])
                rc2.add_clause([-Y[i, p], Y[i, q], B[p, q, 1, 0]])

    for i in range(num_cells):
        for j in range(num_mutations):
            # 0->1
            if alpha == 0:
                if I_mtr[i, j] == 0:
                    rc2.add_clause([-Y[i, j]], weight=1)
                if I_mtr[i, j] == 1:
                    rc2.add_clause([Y[i, j]])

            # 0->1 and 1->0
            if alpha > 0:
                if I_mtr[i, j] == 0:
                    rc2.add_clause([Y[i, j], Z[i, j]])
                    rc2.add_clause([-Y[i, j], -Z[i, j]])
                    rc2.add_clause([Z[i, j]], weight=math.log((1 - alpha) / beta))
                if I_mtr[i, j] == 1:
                    rc2.add_clause([Y[i, j]], weight=math.log((1 - beta) / alpha))

    if kmax > 0:
        for combo in combinations(range(num_mutations), kmax + 1):
            tmp = []
            for com in combo:
                tmp.append(-K[com])
            rc2.add_clause(tmp)

        for j in range(num_mutations):
            type(j)

    type(delta)

    s_time = time.time()
    variables = rc2.compute()
    e_time = time.time()
    running_time = e_time - s_time

    sol_Y = np.empty((num_cells, num_mutations), dtype=np.int8)
    numVar = 0
    for i in range(num_cells):
        for j in range(num_mutations):
            sol_Y[i, j] = variables[numVar] > 0
            numVar += 1

    removedMutsIDs = []
    if kmax > 0:
        sol_K = np.empty(num_mutations, dtype=np.int8)
        numVar = numVarY + numVarB + numVarZ
        for j in range(num_mutations):
            sol_K[j] = variables[numVar] > 0
            numVar += 1
            if sol_K[j]:
                removedMutsIDs.append(snvs[j])

    df_output = pd.DataFrame(sol_Y)
    df_output.columns = snvs
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"

    df_output[removedMutsIDs] = 0

    scp.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output


def phiscsi_bulk(
    df_input,
    alpha,
    beta,
    kmax=0,
    vaf_info=None,
    delta=0.2,
    time_limit=86400,
    n_threads=1,
):
    """Solving using PhISCS-I (in single-cell mode with bulk and mutation elimination).

    a combinatorial approach for subperfect tumor phylogeny reconstruction
    via integrative use of single-cell and bulk sequencing data :cite:`PhISCS`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing
        entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    kmax : :obj:`int`, optional
        Max number of mutations to be eliminated, by default 0
    vaf_info : :class:`pandas.DataFrame`, optional
        Information about the variant allele frequency in bulk data
        The size is n_SNVs x n_samples, by default None
    delta : :obj:`float`, optional
        Delta parameter accounting for VAF variance, by default 0.2
    time_limit : :obj:`int`, optional
        Time limit of the Gurobi running in seconds, by default 86400 (one day)
    n_threads : :obj:`int`, optional
        Number of threads for Gurobi solver, by default 1

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).

    Examples
    --------
    >>> adata = scp.datasets.acute_lymphocytic_leukemia2()
    >>> adata.var["VAF"] = (
        2
        * adata.var["MutantCount"]
        / (adata.var["MutantCount"] + adata.var["ReferenceCount"])
    )
    >>> df_out = scp.tl.phiscsi_bulk(
        adata.to_df(),
        alpha=0.001,
        beta=0.181749,
        delta=0.2,
        kmax=3,
        vaf_info=adata.var[["VAF"]],
    )
    """
    gp, gp_is_not_imported = scp.ul.import_gurobi()
    if gp_is_not_imported:
        scp.logg.error("Unable to import a package!")

    scp.logg.info(
        f"running PhISCS-I-orig with alpha={alpha}, beta={beta}, kmax={kmax}, "
        f"vaf_info={vaf_info}, delta={delta}, time_limit={time_limit}, "
        f"n_threads={n_threads}"
    )

    cells = list(df_input.index)
    snvs = list(df_input.columns)
    df_input = df_input.replace("?", 3)
    df_input = df_input.astype(int)
    I_mtr = df_input.values

    model = gp.Model("ILP")
    model.Params.OutputFlag = 0
    model.Params.LogFile = ""
    model.Params.Threads = n_threads
    model.Params.TimeLimit = time_limit

    numCells = len(cells)
    numMutations = len(snvs)
    if vaf_info is not None:
        vaf_info = vaf_info.loc[snvs]
        sampleIDs = vaf_info.columns
        vaf_info.loc["NULL"] = vaf_info.shape[1] * [1]

    # Matrix Y is matrix of corrected (i.e. true) genotypes w.r.t. input SC matrix I
    Y = {}
    for c in range(numCells):
        for m in range(numMutations):
            Y[c, m] = model.addVar(vtype=gp.GRB.BINARY, name=f"Y({c},{m})")

    # Variables B control the existence of conflict between columns
    B = {}
    for p in range(numMutations + 1):
        for q in range(numMutations + 1):
            B[p, q, 1, 1] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},1,1]"
            )
            B[p, q, 1, 0] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},1,0]"
            )
            B[p, q, 0, 1] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},0,1]"
            )

    K = {}
    for m in range(numMutations + 1):
        K[m] = model.addVar(vtype=gp.GRB.BINARY, name=f"K[{m}]")
    model.addConstr(K[numMutations] == 0)  # null mutation can not be eliminated

    # A[p,q] = 1 if p is ancestor of q
    A = {}
    if vaf_info is not None:
        for p in range(
            numMutations + 1
        ):  # mutation with index numMutation is null mutation
            for q in range(numMutations + 1):
                A[p, q] = model.addVar(vtype=gp.GRB.BINARY, obj=0, name=f"A[{p},{q}]")

    model.update()

    # number of eliminated columns is upper bounded by user provided constant
    model.addConstr(gp.quicksum(K[m] for m in range(numMutations)) <= kmax)

    # Enforce three gametes rule
    for i in range(numCells):
        for p in range(numMutations):
            for q in range(numMutations):
                model.addConstr(Y[i, p] + Y[i, q] - B[p, q, 1, 1] <= 1)
                model.addConstr(-Y[i, p] + Y[i, q] - B[p, q, 0, 1] <= 0)
                model.addConstr(Y[i, p] - Y[i, q] - B[p, q, 1, 0] <= 0)

    # Null mutation present in each cell
    for p in range(numMutations + 1):
        model.addConstr(B[p, numMutations, 1, 0] == 0)

    # Forbid conflict between columns (three gametes rule)
    for p in range(numMutations):
        for q in range(numMutations):
            model.addConstr(
                B[p, q, 0, 1] + B[p, q, 1, 0] + B[p, q, 1, 1] <= 2 + K[p] + K[q]
            )

    # Constraints for integrating VAF obtained from bulk data into the model
    if vaf_info is not None:
        for p in range(numMutations):
            for q in range(p + 1, numMutations):
                model.addConstr(A[p, q] + A[q, p] <= 1 - K[p])
                model.addConstr(A[p, q] + A[q, p] <= 1 - K[q])
                model.addConstr(A[p, q] + A[q, p] >= B[p, q, 1, 1] - K[p] - K[q])
        for p in range(numMutations + 1):
            model.addConstr(A[p, p] == 0)
            for q in range(numMutations + 1):
                model.addConstr(A[p, q] <= 1 - K[p])
                model.addConstr(A[p, q] <= 1 - K[q])

                if p < q:
                    model.addConstr(A[p, q] + A[q, p] <= 1)

                model.addConstr(A[p, q] + B[p, q, 0, 1] <= 1 + K[p] + K[q])
                model.addConstr(
                    B[p, q, 1, 0] + B[p, q, 1, 1] - A[p, q] <= 1 + K[p] + K[q]
                )

                for sampleID in sampleIDs:
                    VAF_p = float(vaf_info.iloc[p][sampleID])
                    VAF_q = float(vaf_info.iloc[q][sampleID])
                    model.addConstr(A[p, q] * VAF_p * (1 + delta) >= A[p, q] * VAF_q)

                    for r in range(numMutations + 1):
                        if r == q:
                            continue
                        VAF_r = float(vaf_info.iloc[r][sampleID])
                        # Constraint 2
                        model.addConstr(
                            VAF_p * (1 + delta)
                            >= VAF_q * (A[p, q] - A[r, q] - A[q, r])
                            + VAF_r * (A[p, r] - A[r, q] - A[q, r])
                        )

                for r in range(numMutations + 1):
                    if r == q:
                        continue
                    # Constraint 1.d
                    model.addConstr(A[p, r] >= A[p, q] + A[q, r] - 1)

            candidateAncestors = list(range(numMutations + 1))
            candidateAncestors.remove(p)

            if p < numMutations:
                model.addConstr(
                    gp.quicksum(A[s, p] for s in candidateAncestors) >= 1 - K[p]
                )
            elif p == numMutations:
                model.addConstr(gp.quicksum(A[s, p] for s in candidateAncestors) == 0)
            else:
                scp.logg.error("p index is out of range")

    # Defining the objective function
    objective = 0
    for j in range(numMutations):
        numZeros = 0
        numOnes = 0
        for i in range(numCells):
            if I_mtr[i][j] == 0:
                numZeros += 1
                objective += np.log(beta / (1 - alpha)) * Y[i, j]
            elif I_mtr[i][j] == 1:
                numOnes += 1
                objective += np.log((1 - beta) / alpha) * Y[i, j]

        objective += numZeros * np.log(1 - alpha)
        objective += numOnes * np.log(alpha)
        objective -= K[j] * (
            numZeros * np.log(1 - alpha)
            + numOnes * (np.log(alpha) + np.log((1 - beta) / alpha))
        )

    model.setObjective(objective, gp.GRB.MAXIMIZE)

    s_time = time.time()
    model.optimize()
    e_time = time.time()
    running_time = e_time - s_time

    removedMutsIDs = []
    sol_K = []
    for j in range(numMutations):
        sol_K.append(K[j].X > 0.5)
        if sol_K[j] == 1:
            removedMutsIDs.append(snvs[j])

    sol_Y = np.zeros((numCells, numMutations), dtype=np.int8)
    for i in range(numCells):
        for j in range(numMutations):
            sol_Y[i, j] = Y[i, j].X > 0.5

    df_output = pd.DataFrame(sol_Y)
    df_output.columns = snvs
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"

    df_output[removedMutsIDs] = 0
    # df_output.drop(removedMutsIDs, axis=1, inplace=True)

    scp.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output


def phiscs_readcount(adata, alpha, beta, time_limit=86400, n_threads=1):
    """Solving using PhISCS-ReadCount (only in single-cell mode, no bulk).

    a combinatorial approach for subperfect tumor phylogeny reconstruction
    via integrative use of single-cell and bulk sequencing data :cite:`PhISCS`.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        Input data contains layers of mutant and total.
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    time_limit : :obj:`int`, optional
        Time limit of the Gurobi running in seconds, by default 86400 (one day)
    n_threads : :obj:`int`, optional
        Number of threads for Gurobi solver, by default 1

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    gp, gp_is_not_imported = scp.ul.import_gurobi()
    if gp_is_not_imported:
        scp.logg.error("Unable to import a package!")

    scp.logg.info(
        f"running PhISCS-readcout with alpha={alpha}, beta={beta},"
        f" time_limit={time_limit}"
    )

    PROB_SEQ_ERROR = 0.001
    BETABINOM_ALPHA = 1.0
    BETABINOM_BETA = 1.0

    def _prob_absent(v, t):
        prob = binom.pmf(v, t, PROB_SEQ_ERROR)
        return prob

    def _prob_present(v, t):
        prob = pmf_BetaBinomial(v, t, BETABINOM_ALPHA, BETABINOM_BETA)
        return prob

    df_input = adata.to_df()
    cells = adata.obs_names
    muts = adata.var_names
    T = adata.layers["total"]
    V = adata.layers["mutant"]

    B_mu = np.zeros((len(cells), len(muts)), dtype=np.float64)
    B_delta = np.zeros((len(cells), len(muts)), dtype=np.float64)
    for i in range(len(cells)):
        for j in range(len(muts)):
            if T[i, j] != 0:
                B_mu[i, j] = _prob_absent(V[i, j], T[i, j])
                B_delta[i, j] = _prob_present(V[i, j], T[i, j])

    model = gp.Model("ILP")
    model.Params.OutputFlag = 0
    model.Params.LogFile = ""
    model.Params.Threads = n_threads
    model.Params.TimeLimit = time_limit
    Y = {}
    B = {}
    for c in range(len(cells)):
        for m in range(len(muts)):
            Y[c, m] = model.addVar(vtype=gp.GRB.BINARY, name=f"Y({c},{m})")
    for p in range(len(muts)):
        for q in range(len(muts)):
            B[p, q, 1, 1] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},1,1]"
            )
            B[p, q, 1, 0] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},1,0]"
            )
            B[p, q, 0, 1] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},0,1]"
            )
    for p in range(len(muts)):
        for q in range(len(muts)):
            model.addConstr(B[p, q, 0, 1] + B[p, q, 1, 0] + B[p, q, 1, 1] <= 2)
            for i in range(len(cells)):
                model.addConstr(Y[i, p] + Y[i, q] - B[p, q, 1, 1] <= 1)
                model.addConstr(-Y[i, p] + Y[i, q] - B[p, q, 0, 1] <= 0)
                model.addConstr(Y[i, p] - Y[i, q] - B[p, q, 1, 0] <= 0)
    objective = 0
    for j in range(len(muts)):
        for i in range(len(cells)):
            if T[i, j] != 0:
                objective += (1 - Y[i, j]) * np.log(
                    B_mu[i, j] * (1 - alpha) + B_delta[i, j] * (alpha)
                )
                objective += (Y[i, j]) * np.log(
                    B_mu[i, j] * (beta) + B_delta[i, j] * (1 - beta)
                )
    model.setObjective(objective, gp.GRB.MAXIMIZE)

    s_time = time.time()
    model.optimize()
    e_time = time.time()
    running_time = e_time - s_time

    sol_Y = np.zeros((len(cells), len(muts)), dtype=np.int8)
    for i in range(len(cells)):
        for j in range(len(muts)):
            sol_Y[i, j] = Y[i, j].X > 0.5

    df_output = pd.DataFrame(sol_Y)
    df_output.columns = muts
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"

    scp.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output



import time

import pandas as pd

import scphylo as scp
from scphylo.external.gpps import gpps_hc, gpps_ilp


def gpps(
    df_input,
    alpha,
    beta,
    k_dollo=0,
    max_del=-1,
    neighbor_size=30,
    n_iters=100,
    time_limit=86400,
    n_threads=1,
):
    """Solving using gpps.

    an ILP-based approach for inferring cancer progression with mutation losses from
    single cell data :cite:`gpps`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing
        entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    k_dollo : :obj:`int`, optional
        k for Dollo model, by default 0
    max_del : :obj:`int`, optional
        Maximum number of deletion allowed, by default -1
    neighbor_size : :obj:`int`, optional
        Hill climbing neighborhood size, by default 30
    n_iters : :obj:`int`, optional
        Hill climbing maximum iterations, by default 100
    time_limit : :obj:`int`, optional
        Time limit (in seconds), by default 86400
    n_threads : :obj:`int`, optional
        Number of threads, by default 1

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    scp.logg.info(
        f"running gpps with alpha={alpha}, beta={beta}, k_dollo={k_dollo}, "
        f"max_del={max_del}, neighbor_size={neighbor_size}, n_iters={n_iters}, "
        f"time_limit={time_limit}, n_threads={n_threads}"
    )

    cells = list(df_input.index)
    snvs = list(df_input.columns)

    s_time = time.time()
    ilp_matrix = gpps_ilp(
        df_input.values,
        alpha=beta,  # gpps takes a as false-negative and b as false-positive
        beta=alpha,
        k_dollo=k_dollo,
        max_del=max_del,
        time_limit=time_limit,
        n_threads=n_threads,
    )
    ilp_matrix = pd.DataFrame(ilp_matrix)

    output_matrix = gpps_hc(
        df_input.values,
        ilp_matrix,
        alpha=beta,
        beta=alpha,
        k_dollo=k_dollo,
        mut_names=snvs,
        ns=neighbor_size,
        mi=n_iters,
    )
    e_time = time.time()
    running_time = e_time - s_time

    df_output = pd.DataFrame(output_matrix)
    df_output.columns = snvs
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"

    scp.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output



import os

import click
import numpy as np
from joblib import Parallel, delayed
from tqdm import tqdm

import scphylo as scp

# from scphylo.pl._trees import _newick_info2_mutation_list


def run_scistree(df_in, alpha, beta, outfile):
    type(outfile)
    # scp.settings.logfile = f"{outfile}/fn_{beta}-fp_{alpha}.log"
    scp.settings.verbosity = "error"
    df_out = scp.tl.scistree(df_in, alpha, beta)
    nll = scp.ul.calc_nll_matrix(df_in, df_out, alpha, beta)
    # scp.io.write(df_out, f"{outfile}/fn_{beta}-fp_{alpha}.CFMatrix")

    # tree = scp.ul.to_tree(df_out)
    # newick, info2, _ = _newick_info2_mutation_list(tree)
    # with open(f"{outfile}/fn_{beta}-fp_{alpha}.newick", "w") as fout:
    #     fout.write(newick + "\n")
    # info2.to_csv(f"{outfile}/fn_{beta}-fp_{alpha}.info2", index=None)
    return nll, df_out, alpha, beta


@click.command(short_help="Grid search for all parameters.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.option(
    "--n_threads",
    "-p",
    default=-1,
    type=int,
    show_default=True,
    help="Number of threads.",
)
def search(genotype_file, n_threads):
    """Grid search for all parameters of alpha and beta.

    scphylo search input.SC
    """
    outfile = os.path.splitext(genotype_file)[0]
    # scp.ul.mkdir(outfile)

    df_in = scp.io.read(genotype_file)

    betas = [0.1, 0.2, 0.3, 0.4]
    alphas = [0.1, 0.01, 0.001, 0.0001, 0.00001]
    n_samples = len(betas) * len(alphas)
    if n_threads == -1:
        n_threads = n_samples

    with scp.ul.tqdm_joblib(
        tqdm(
            ascii=True,
            ncols=100,
            desc="GRID SEARCHING:",
            total=n_samples,
            position=0,
        )
    ):
        output = Parallel(n_jobs=n_threads)(
            delayed(run_scistree)(df_in, alpha, beta, outfile)
            for alpha in alphas
            for beta in betas
        )
        output_i = np.argmin([x[0] for x in output])
        df_out = output[output_i][1]
        alpha = output[output_i][2]
        beta = output[output_i][3]
        scp.io.write(df_out, f"{outfile}.scistree.CFMatrix")
        scp.settings.verbosity = "info"
        scp.settings.logfile = f"{outfile}.scistree.log"
        scp.logg.info(f"running ScisTree with alpha={alpha}, beta={beta}, n_threads=1")
        scp.ul.stat(df_in, df_out, alpha, beta, 0)

    return None



import time

import apted
import numpy as np

import scphylo as scp
from scphylo.external._mltd import run_mltd
from scphylo.ul._trees import _split_labels, _to_apted


def gs(df_grnd, df_sol):
    """Genotype-similarity accuracy.

    This measure was introduced in :cite:`SiCloneFit`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        Similarity out of one.
    """
    muts = np.intersect1d(df_grnd.columns, df_sol.columns)
    cells = np.intersect1d(df_grnd.index, df_sol.index)
    if len(muts) == 0:
        scp.logg.error("No common mutations found between two trees!")
    if len(cells) == 0:
        scp.logg.error("No common cells found between two trees!")
    M_grnd = df_grnd.loc[cells, muts].values
    M_sol = df_sol.loc[cells, muts].values
    return 1 - np.abs(M_grnd - M_sol).sum() / M_grnd.size


def ad(df_grnd, df_sol):
    """Ancestor-descendent accuracy.

    For each pair of mutations in ground truth tree that are in
    ancestor-descendant relation (same nodes excluded) we check
    whether this relationship is preserved in the inferred tree.

    This measure was introduced in :cite:`B-SCITE`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        Similarity out of one.
    """
    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    if len(inter) == 0:
        scp.logg.error("No common mutations found between two trees!")
    M_grnd = df_grnd[inter].values
    M_sol = df_sol[inter].values
    error_pairs = []
    n_adpairs = 0
    for i in range(M_grnd.shape[1]):
        for j in range(i, M_grnd.shape[1]):
            cap1 = M_grnd[:, i] * M_grnd[:, j]
            cap2 = M_sol[:, i] * M_sol[:, j]
            if np.sum(cap1) > 0 and np.sum(M_grnd[:, i]) != np.sum(M_grnd[:, j]):
                n_adpairs = n_adpairs + 1
                if np.sum(cap2) == 0:
                    error_pairs.append([i, j])
                else:
                    if np.sum(M_grnd[:, j]) > np.sum(M_grnd[:, i]) and np.sum(
                        M_sol[:, j]
                    ) <= np.sum(M_sol[:, i]):
                        error_pairs.append([i, j])
                    else:
                        if np.sum(M_grnd[:, i]) > np.sum(M_grnd[:, j]) and np.sum(
                            M_sol[:, i]
                        ) <= np.sum(M_sol[:, j]):
                            error_pairs.append([i, j])
    if n_adpairs == 0:
        scp.logg.error("No pair of mutations exists in the given inputs!")
    return 1 - len(error_pairs) / n_adpairs


def dl(df_grnd, df_sol):
    """Different-lineage accuracy.

    For each pair of mutations in ground truth tree that are
    in different-lineages relation we check whether the same relationship
    is preserved in the inferred tree.

    This measure was introduced in :cite:`B-SCITE`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        Similarity out of one.
    """
    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    if len(inter) == 0:
        scp.logg.error("No common mutations found between two trees!")
    M_grnd = df_grnd[inter].values
    M_sol = df_sol[inter].values
    n_dlpairs1 = 0
    n_dlpairs2 = 0
    for i in range(M_grnd.shape[1]):
        for j in range(i, M_grnd.shape[1]):
            cap1 = M_grnd[:, i] * M_grnd[:, j]
            cap2 = M_sol[:, i] * M_sol[:, j]
            if (
                np.sum(cap1) == 0
                and np.sum(M_grnd[:, i]) != 0
                and np.sum(M_grnd[:, j]) != 0
            ):
                n_dlpairs1 = n_dlpairs1 + 1
                if (
                    np.sum(cap2) == 0
                    and np.sum(M_sol[:, i]) != 0
                    and np.sum(M_sol[:, j]) != 0
                ):
                    n_dlpairs2 = n_dlpairs2 + 1
    return n_dlpairs2 / n_dlpairs1


def cc(df_grnd, df_sol):
    """Co-clustering accuracy.

    For each pair of mutations in ground truth tree that are on the same node we look
    relationship is preserved in the inferred tree.

    This measure was introduced in :cite:`B-SCITE`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        Similarity out of one.
    """
    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    if len(inter) == 0:
        scp.logg.error("No common mutations found between two trees!")
    M_grnd = df_grnd[inter].values
    M_sol = df_sol[inter].values
    type(M_grnd)
    type(M_sol)

    # TODO: implement
    return None


def mltd(df_grnd, df_sol):
    """Multi-labeled tree dissimilarity measure (MLTD).

    This measure was introduced in :cite:`MLTD`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`dict`
        {'distance', 'similarity', 'normalized_similarity'}
    """

    def _convert_tree_to_mtld_input(tree, file):
        with open(file, "w") as fout:
            for u, v, l in tree.edges.data("label"):
                if tree.in_degree(u) == 0:
                    fout.write(f"{u}=\n")
                muts = l.split(tree.graph["splitter_mut"])
                fout.write(f"{v}={','.join(muts)}\n")
            for u in tree.nodes:
                children = [str(n) for n in tree.neighbors(u)]
                if len(children) == 0:
                    continue
                fout.write(f"{u}:{','.join(children)}\n")

    tmpdir = scp.ul.tmpdirsys(suffix=".mltd")

    df_grnd.columns = df_grnd.columns.str.replace(":", "_").str.replace("=", "_")
    df_sol.columns = df_sol.columns.str.replace(":", "_").str.replace("=", "_")
    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    if len(inter) == 0:
        scp.logg.error("No common mutations found between two trees!")
    df_grnd1 = df_grnd[inter]
    df_sol1 = df_sol[inter]

    tree_grnd = scp.ul.to_tree(df_grnd1)
    tree_sol = scp.ul.to_tree(df_sol1)

    _convert_tree_to_mtld_input(tree_grnd, f"{tmpdir.name}/grnd.in")
    _convert_tree_to_mtld_input(tree_sol, f"{tmpdir.name}/sol.in")

    s_time = time.time()
    result = run_mltd(f"{tmpdir.name}/grnd.in", f"{tmpdir.name}/sol.in")
    e_time = time.time()
    running_time = e_time - s_time
    type(running_time)

    tmpdir.cleanup()
    if result is None:
        scp.logg.error("MLTD core failed!")

    return result


def tpted(df_grnd, df_sol):
    """Tumor phylogeny tree edit distance measure (TPTED).

    This measure was introduced in :cite:`PhISCS`. This implementation uses
    `APTED <https://github.com/JoaoFelipe/apted>`_.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        Similarity out of one.
    """
    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    if len(inter) == 0:
        scp.logg.error("No common mutations found between two trees!")
    df_grnd1 = df_grnd[inter]
    df_sol1 = df_sol[inter]

    tree_grnd = scp.ul.to_tree(df_grnd1)
    tree_sol = scp.ul.to_tree(df_sol1)
    inter = np.setdiff1d(
        inter,
        np.union1d(
            tree_grnd.graph["become_germline"], tree_sol.graph["become_germline"]
        ),
    )

    df_grnd1 = df_grnd[inter]
    df_sol1 = df_sol[inter]

    tree_grnd = scp.ul.to_tree(df_grnd1)
    tree_sol = scp.ul.to_tree(df_sol1)

    mt_grnd = scp.ul.to_mtree(tree_grnd)
    mt_sol = scp.ul.to_mtree(tree_sol)

    sl_grnd, sl_sol = _split_labels(mt_grnd, mt_sol)

    apted_grnd = _to_apted(sl_grnd)
    apted_sol = _to_apted(sl_sol)

    tree1 = apted.helpers.Tree.from_text(apted_grnd)
    tree2 = apted.helpers.Tree.from_text(apted_sol)

    ap = apted.APTED(tree1, tree2)
    ed = ap.compute_edit_distance()

    # FIXME: `python -m apted -t {a{b}{c}} {a{c}{b}}` returns 2!
    # looks like the above isomorphic trees have TED of 2!
    return 1 - ed / (2 * (len(inter) + 1))



"""Settings Module."""

from scphylo.settings._settings import (
    hg19,
    hg38,
    logfile,
    m10x,
    mm10,
    refs,
    tools,
    verbosity,
)

__all__ = (
    hg19,
    hg38,
    logfile,
    m10x,
    mm10,
    refs,
    tools,
    verbosity,
)



import scphylo as scp


def cardelino(adata, mode, n_clones=None, tree=None):
    """Cardelino.

    Computational integration of somatic clonal substructure and single-cell
    transcriptomes :cite:`Cardelino`.

    This method uses the prebuilt tree (either by bWES or scDNA) and
    then tries to cluster single-cells of scRNA based on the appearance
    of mutations detect in the tree.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        Input data contains layers of mutant and total.
    mode : :obj:`str`
        Cardelino mode, possible values are:

            - `fixed`: in this mode provide `tree`
            - `free`: in this mode provide `n_clones`
    n_clones : :obj:`int`, optional
        Number of clones, by default None
    tree : :class:`pandas.DataFrame`, optional
        Input prebuilt tree in the form of conflict-free matrix, by default None

    Returns
    -------
    :class:`pandas.DataFrame`
        The assignment of single-cells to the clones of the tree.
    """
    cardelino, cardelino_is_not_imported = scp.ul.import_rpy2(
        "cardelino",
        "devtools::install_github('single-cell-genetics/cardelino'\n",
    )
    if cardelino_is_not_imported:
        scp.logg.error("Unable to import a package!")

    import rpy2.robjects as ro
    from rpy2.robjects import pandas2ri

    scp.logg.info("running Cardelino")

    config = None
    with ro.conversion.localconverter(ro.default_converter + pandas2ri.converter):
        A = ro.conversion.py2rpy(adata.to_df(layer="mutant").T)
        D = ro.conversion.py2rpy(adata.to_df(layer="total").T)
        if mode == "fixed":
            config = ro.conversion.py2rpy(tree.T)

    ro.globalenv["A"] = A
    ro.globalenv["D"] = D
    if mode == "fixed":
        ro.globalenv["config"] = config

    if mode == "free":
        assignments = cardelino.clone_id(A, D, n_clone=n_clones)
    elif mode == "fixed":
        assignments = cardelino.clone_id(A, D, Config=config)
    else:
        scp.logg.error("Wrong mode")

    result = cardelino.assign_cells_to_clones(assignments.rx2["prob"])

    with ro.conversion.localconverter(ro.default_converter + pandas2ri.converter):
        result = ro.conversion.rpy2py(result)

    return result



__author__ = "Simone Ciccolella"
__date__ = "11/30/21"


def expand_name(s, max_gains, max_losses):
    positive_names = [s + "+" + str(i) for i in range(max_gains)]
    negative_names = [s + "-" + str(i) for i in range(max_losses)]
    return positive_names + negative_names



import os

import click

import scphylo as scp


@click.command(short_help="Run PhISCS (CSP version).")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
def phiscsb(genotype_file, alpha, beta):
    """PhISCS-B.

    A combinatorial approach for subperfect
    tumor phylogeny reconstructionvia integrative use of
    single-cell and bulk sequencing data :cite:`PhISCS`.

    scphylo solver phiscsb input.SC 0.0001 0.1
    """
    outfile = os.path.splitext(genotype_file)[0]

    scp.settings.verbosity = "info"
    scp.settings.logfile = f"{outfile}.phiscsb.log"

    df_in = scp.io.read(genotype_file)
    df_out = scp.tl.phiscsb(df_in, alpha=alpha, beta=beta)
    scp.io.write(df_out, f"{outfile}.phiscsb.CFMatrix")

    return None


@click.command(short_help="Run PhISCS (ILP version).")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
@click.option(
    "--time_limit",
    "-t",
    default=86400,
    type=int,
    show_default=True,
    help="Time limit of the program (in second).",
)
@click.option(
    "--n_threads",
    "-p",
    default=1,
    type=int,
    show_default=True,
    help="Number of threads.",
)
def phiscsi(genotype_file, alpha, beta, time_limit, n_threads):
    """PhISCS-I.

    A combinatorial approach for subperfect
    tumor phylogeny reconstructionvia integrative use of
    single-cell and bulk sequencing data :cite:`PhISCS`.

    scphylo phiscsi input.SC 0.0001 0.1 -t 3600 -p 8
    """
    outfile = os.path.splitext(genotype_file)[0]

    scp.settings.verbosity = "info"
    scp.settings.logfile = f"{outfile}.phiscsi.log"

    df_in = scp.io.read(genotype_file)
    df_out = scp.tl.phiscsi(
        df_in, alpha=alpha, beta=beta, time_limit=time_limit, n_threads=n_threads
    )
    scp.io.write(df_out, f"{outfile}.phiscsi.CFMatrix")

    return None



import glob
import subprocess

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run GATK.")
@click.argument(
    "outdir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.argument(
    "dna_or_rna",
    required=True,
    type=click.Choice(["dna", "rna"]),
)
@click.option(
    "--time",
    default="0-20:00:00",
    type=str,
    show_default=True,
    help="Time.",
)
@click.option(
    "--mem",
    default="90",
    type=str,
    show_default=True,
    help="Memory.",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def gatk(outdir, ref, dna_or_rna, time, mem, afterok):
    """Run GATK.

    caller caller gatk /path/to/in/dir hg19|hg38|mm10 dna|rna

    BAM files (*.mapped.bam) --> BAM files (*.markdup_bqsr.bam)
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    def get_command(sample):
        cmds = ""
        cmds += cmd([f"module load {scp.settings.tools['gatk']}"])
        cmds += cmd(
            [
                f'gatk --java-options "-Xmx{int(mem)-10}g"',
                "AddOrReplaceReadGroups",
                f"--INPUT {outdir}/{sample}.mapped.bam",
                f"--OUTPUT {outdir}/{sample}.rgadded.bam",
                "--SORT_ORDER coordinate",
                f"--RGID {sample}",
                "--RGLB lib1",
                "--RGPL ILLUMINA",
                "--RGPU unit1",
                f"--RGSM {sample}",
            ]
        )
        cmds += cmd(
            [
                f'gatk --java-options "-Xmx{int(mem)-10}g"',
                "MarkDuplicates",
                f"--INPUT {outdir}/{sample}.rgadded.bam",
                f"--OUTPUT {outdir}/{sample}.markdup.bam",
                f"--METRICS_FILE {outdir}/{sample}.metrics",
            ]
        )
        cmds += cmd(
            [
                "rm -rf",
                f"{outdir}/{sample}.rgadded.bam",
                f"{outdir}/{sample}.metrics",
            ]
        )
        if dna_or_rna == "rna":
            cmds += cmd(
                [
                    f'gatk --java-options "-Xmx{int(mem)-10}g"',
                    "SplitNCigarReads",
                    f"--reference {config['ref']}",
                    f"--input {outdir}/{sample}.markdup.bam",
                    f"--output {outdir}/{sample}.splitted.bam",
                ]
            )
        cmds += cmd(
            [
                f'gatk --java-options "-Xmx{int(mem)-10}g"',
                "BaseRecalibrator",
                f"--reference {config['ref']}",
                (
                    f"--input {outdir}/{sample}.splitted.bam"
                    if dna_or_rna == "rna"
                    else f"--input {outdir}/{sample}.markdup.bam"
                ),
                f"--output {outdir}/{sample}.markdup_bqsr.report",
                "--known-sites",
                f"{' --known-sites '.join(config['known_sites'])}",
            ]
        )
        cmds += cmd(
            [
                f'gatk --java-options "-Xmx{int(mem)-10}g"',
                "ApplyBQSR",
                f"--reference {config['ref']}",
                f"--bqsr-recal-file {outdir}/{sample}.markdup_bqsr.report",
                (
                    f"--input {outdir}/{sample}.splitted.bam"
                    if dna_or_rna == "rna"
                    else f"--input {outdir}/{sample}.markdup.bam"
                ),
                f"--output {outdir}/{sample}.markdup_bqsr.bam",
            ]
        )
        cmds += cmd(
            [
                "rm -rf",
                f"{outdir}/{sample}.markdup.bam",
                f"{outdir}/{sample}.markdup_bqsr.report",
            ]
        )
        if dna_or_rna == "rna":
            cmds += cmd(
                [
                    "rm -rf",
                    f"{outdir}/{sample}.splitted.bam",
                    f"{outdir}/{sample}.splitted.bai",
                ]
            )
        cmds += cmd(["echo Done!"], islast=True)
        return cmds

    files = glob.glob(f"{outdir}/*.mapped.bam")
    temp = []
    for file in files:
        file = file.split("/")[-1].replace(".mapped.bam", "")
        temp.append({"sample": file})
    df_cmds = pd.DataFrame(temp)
    df_cmds["cmd"] = df_cmds.apply(lambda x: get_command(x["sample"]), axis=1)

    cmdmain = write_cmds_get_main(
        df_cmds,
        "gatk",
        time,
        mem,
        None,
        1,
        scp.settings.tools["email"],
        f"{outdir}/_tmp",
        afterok,
    )
    code = subprocess.getoutput(cmdmain)
    scp.logg.info(code)

    return None



import os
import time

import scphylo as scp


def cellphy(adata, mode="fast"):
    """Solving using CellPhy.

    Accurate and fast probabilistic inference of single-cell phylogenies
    from scDNA-seq data :cite:`CellPhy`.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        Input data contains layers of mutant and total.
    mode : :obj:`str`
        CellPhy mode, possible values are:

            - `fast`
            - `full`

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    executable = scp.ul.executable("cellphy.sh", "CellPhy")

    if adata.shape[1] < 37:
        scp.logg.error("number of mutations must be greater than 37!")

    scp.logg.info(f"running CellPhy on {mode} mode")

    tmpdir = scp.ul.tmpdirsys(suffix=".cellphy")

    scp.io.to_vcf(adata, filepath=f"{tmpdir.name}/cellphy.vcf")

    cmd = f"{executable} {mode.upper()} {tmpdir.name}/cellphy.vcf"

    s_time = time.time()
    os.system(cmd)
    e_time = time.time()
    running_time = e_time - s_time
    type(running_time)

    with open(f"{tmpdir.name}/cellphy.vcf.raxml.bestTree") as fin:
        tree = fin.readline().strip()

    tmpdir.cleanup()

    return tree



import os

import anndata as ad
import ete3
import networkx as nx
import numpy as np
import pandas as pd

import scphylo as scp


def read(filepath):
    """Read genotype matrix and read-count matrix.

    The genotype matrix must be in the in format of :class:`pandas.DataFrame`
    The read-count matrix must be in the format of :class:`anndata.AnnData`.

    Parameters
    ----------
    filepath : :obj:`str`
        The path to the file. The extension must be one of
        [`.tsv`, `.SC`, `.CFMatrix`, `.h5ad`, `.h5ad.gz`, `.nwk`]

    Returns
    -------
    :class:`pandas.DataFrame` or :class:`anndata.AnnData`
        Depends on the format of the input file the output type is different.
    """
    ext = os.path.splitext(filepath)[-1]
    if ext in [".SC", ".CFMatrix", ".before_FP_FN_NA", ".tsv"]:
        sc = pd.read_table(filepath, index_col=0)
        if len(sc.columns) != len(set(sc.columns)):
            scp.logg.error("Mutation ids must be unique!")
        return sc
    elif ext in [".csv"]:
        sc = pd.read_csv(filepath, index_col=0)
        if len(sc.columns) != len(set(sc.columns)):
            scp.logg.error("Mutation ids must be unique!")
        return sc
    elif ext in [".h5ad", ".gz"]:
        return ad.read_h5ad(filepath)
    elif ext in [".nwk"]:
        return _read_nwk(filepath)
    else:
        scp.logg.error("Extension is wrong!")
        return None


def write(obj, filepath):
    """Write genotype matrix or read-count matrix into a file.

    Parameters
    ----------
    obj : :class:`pandas.DataFrame` or :class:`anndata.AnnData`
        The input object which is going to be written in a file.
    filepath : :obj:`str`
        The file path where the `obj` must be written in.
    """
    if isinstance(obj, pd.DataFrame):
        obj.index.name = "cellIDxmutID"
        obj.to_csv(filepath, sep="\t")
    elif isinstance(obj, ad.AnnData):
        obj.write(filepath + ".h5ad.gz", compression="gzip")
    else:
        scp.logg.error("Object instance is wrong!")


def _read_nwk(filepath):
    tree = ete3.Tree(filepath, format=1)
    G = nx.DiGraph()
    node2id = {}
    i = 0
    for n in tree.traverse("postorder"):
        if n.name == "" or "Inner" in n.name:
            G.add_node(i, label="––")
        else:
            G.add_node(i, label=str(n.name))
        node2id[n] = i
        i += 1

    for p in tree.traverse("postorder"):
        pn = node2id[p]
        for c in p.children:
            cn = node2id[c]
            G.add_edge(pn, cn)

    i = 0
    for e, u, _ in G.edges.data("label"):
        G.edges[(e, u)]["label"] = f"m{i}"
        i += 1
    G.graph["normal_cells"] = []
    G.graph["splitter_mut"] = "\n"
    G.graph["splitter_cell"] = "\n"
    data = scp.ul.to_cfmatrix(G)
    return data


def to_vcf(adata, filepath):
    def _calculate_pl(ref_count, alt_count):
        if ref_count == 0 and alt_count == 0:
            return [0, 0, 0]
        elif ref_count == 0:
            return [0, 0, 255]
        elif alt_count == 0:
            return [255, 0, 0]
        else:
            total_count = ref_count + alt_count
            p_ref = ref_count / total_count
            p_alt = alt_count / total_count
            p_ref_given_data = p_ref
            p_alt_given_data = p_alt
            p_ref_ref_given_data = p_ref * p_ref
            pl_ref = -10 * np.math.log10(p_ref_given_data)
            pl_alt = -10 * np.math.log10(p_alt_given_data)
            pl_ref_ref = -10 * np.math.log10(p_ref_ref_given_data)
            return [int(pl_ref), int(pl_ref_ref), int(pl_alt)]

    gt_mtx = np.where(
        adata.X == 0,
        "0/0",
        np.where(adata.X == 1, "0/1", np.where(adata.X == 3, "./.", adata.X)),
    )
    ref_mtx = adata.layers["total"] - adata.layers["mutant"]
    alt_mtx = adata.layers["mutant"]
    with open(filepath, "w") as fout:
        fout.write("##fileformat=VCFv4.3\n")
        fout.write("##contig=<ID=chr1,length=1000000>\n")
        fout.write('##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">\n')
        fout.write(
            '##FORMAT=<ID=PL,Number=G,Type=Integer,Description="Normalized,'
            " Phred-scaled likelihoods for genotypes as defined in the VCF"
            ' specification">\n'
        )
        fout.write("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t")
        fout.write("\t".join(adata.obs_names) + "\n")
        for j in range(adata.shape[1]):
            samples = []
            for i in range(adata.shape[0]):
                pl = _calculate_pl(ref_mtx[i, j], alt_mtx[i, j])
                samples.append(gt_mtx[i, j] + ":" + ",".join(map(str, pl)))
            variant = (
                "\t".join(
                    [
                        "chr1",
                        str(j + 1),
                        ".",
                        "A",
                        "C",
                        ".",
                        "PASS",
                        ".",
                        "GT:PL",
                        "\t".join(samples),
                    ]
                )
                + "\n"
            )
            fout.write(variant)



import sys

__author__ = "Simone Ciccolella"
__date__ = "11/30/21"

name = "nh2lgf"

desc = """

convert newhampshire (or newick) format (nh)
to lemon graph format (lgf)
-- Murray Patterson, Sept 2014

"""

note = ""


# for processing a 'node' object, i,e., of the form 'A:B', where 'A'
# is a name (or a bootstrap value in the case of an internal node),
# and 'B' is a branch length
def node(suffix, count, nodes, arcs):

    assert not suffix.startswith("("), "\n\ninput not in nh format"

    # eat until ',', ')' or ';'
    if ")" not in suffix:  # we know we're at the end
        a, b, c = suffix.partition(";")  # if not, search for ','
    elif "," in suffix and suffix.find(",") < suffix.find(")"):
        a, b, c = suffix.partition(",")
    else:  # o.w., it is the only terminal
        a, b, c = suffix.partition(")")

    name, x, brlen = a.partition(":")
    nodes.append((count + 1, name))

    return b + c, count + 1, brlen, nodes, arcs


# for processing a 'subtree' object, i.e., of the form 'N' or 'X',
# where 'N' is a 'node' object (base case), and 'X' is a 'subtrees'
# object (recursive case)
def subtree(suffix, count, nodes, arcs):

    if suffix.startswith("("):  # recursive case
        return subtrees(suffix, count, nodes, arcs)

    else:  # base case, we're at a node (which is necessarily a leaf, here)
        return node(suffix, count, nodes, arcs)


# for recieving and processing a 'subtrees' object, i.e. of the form
# '(S,S,..,S)' where 'S' is a 'subtree' object
def subtrees(suffix, count, nodes, arcs):

    assert ")" in suffix, "\n\ninput not in nh format"

    suffix = suffix[1:]  # eat the '('
    children = []  # store info for each 'S' in '(S,S,..,S)'

    # loop through the 'S' in '(S,S,..,S)'
    while True:  # note : contains at least one 'S', even if ''
        suffix, count, brlen, nodes, arcs = subtree(suffix, count, nodes, arcs)
        children.append((count, brlen))

        if suffix.startswith(")"):  # stopping condition
            break
        if suffix.startswith(","):  # in case of multiple subtrees
            suffix = suffix[1:]  # eat the ','

    # now, here, suffix[0] == ')'
    suffix = suffix[1:]  # eat the ')'
    # get parent of '(S,S,..,S)'
    suffix, count, brlen, nodes, arcs = node(suffix, count, nodes, arcs)
    for child in children:  # now add the arcs to each child 'S'
        arcs.append((child[0], count, child[1]))

    return suffix, count, brlen, nodes, arcs


def newick_to_edgelist(input_string):
    s = input_string
    assert s.find(";") == len(s) - 1, "\n\nerror: input not in nh format"
    assert s.count("(") == s.count(")"), "\n\nerror: input not in nh format"

    # parse the input
    nodes = []
    arcs = []
    subtree(s, -1, nodes, arcs)

    try:
        # output in lgf format
        node_dict = {}
        for label, name in nodes:
            node_dict[label] = name

        edges = []
        for u, v, _ in arcs:
            edges.append((u, v))

        return node_dict, edges
    except OSError:
        sys.exit(0)



import time

import numpy as np
import pandas as pd

import scphylo as scp


def onconem(df_input, alpha, beta):
    """Solving using OncoNEM.

    Inferring tumor evolution from single-cell sequencing data :cite:`OncoNEM`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix where 0 is absent, 1 is present and 3 is missing.
    alpha : :obj:`float`
        False positive rate.
    beta : :obj:`float`
        False negative rate.

    Returns
    -------
    :class:`pandas.DataFrame`
        Conflict-free genotype matrix where 0 is absent, 1 is present.
    """
    onconem, onconem_is_not_imported = scp.ul.import_rpy2(
        "oncoNEM",
        "BiocManager::install('graph')\n"
        + "devtools::install_bitbucket('edith_ross/oncoNEM')\n",
    )
    if onconem_is_not_imported:
        scp.logg.error("Unable to import a package!")

    import rpy2.robjects as ro
    from rpy2.robjects import numpy2ri
    from rpy2.robjects.packages import importr

    igraph = importr("igraph")

    scp.logg.info(f"running OncoNEM with alpha={alpha}, beta={beta}")

    with ro.conversion.localconverter(ro.default_converter + numpy2ri.converter):
        data = ro.conversion.py2rpy(df_input.replace(3, 2).T.values)

    s_time = time.time()
    onem = onconem.oncoNEM(Data=data, FPR=alpha, FNR=beta)
    ro.globalenv["onem"] = onem
    ro.r("onem$search(delta=200)")
    onem_expanded = onconem.expandOncoNEM(
        onem, epsilon=10, delta=200, checkMax=10000, app=True
    )
    onco_tree = onconem.clusterOncoNEM(oNEM=onem_expanded, epsilon=10)
    igraph.get_edgelist(onco_tree.rx2["g"])
    post = onconem.oncoNEMposteriors(
        tree=onco_tree.rx2["g"],
        clones=onco_tree.rx2["clones"],
        Data=data,
        FPR=alpha,
        FNR=beta,
    )
    sol_Y = np.rint(post.rx2["p_mut"])
    e_time = time.time()
    running_time = e_time - s_time

    df_output = pd.DataFrame(sol_Y)
    df_output = df_output.T
    df_output.columns = df_input.columns
    df_output.index = df_input.index
    df_output.index.name = "cellIDxmutID"
    df_output = df_output.astype(int)

    scp.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output



import os

import click

import scphylo as scp


@click.command(short_help="Run GRMT.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
@click.option(
    "--n_iters",
    "-l",
    default=30,
    type=int,
    show_default=True,
    help="Number of iterations.",
)
@click.option(
    "--n_threads",
    "-p",
    default=1,
    type=int,
    show_default=True,
    help="Number of threads.",
)
def grmt(genotype_file, alpha, beta, n_iters, n_threads):
    """GRMT.

    Generative Reconstruction of Mutation Tree From Scratch Using Single-Cell
    Sequencing Data :cite:`GRMT`.

    scphylo solver grmt input.SC 0.0001 0.1 -l 100 -p 2
    """
    outfile = os.path.splitext(genotype_file)[0]

    scp.settings.verbosity = "info"
    scp.settings.logfile = f"{outfile}.grmt.log"

    df_in = scp.io.read(genotype_file)
    df_out = scp.tl.grmt(
        df_in, alpha=alpha, beta=beta, n_iters=n_iters, n_threads=n_threads
    )
    scp.io.write(df_out, f"{outfile}.grmt.CFMatrix")

    return None



"""Plotting Module."""

from scphylo.pl._data import heatmap
from scphylo.pl._trees import clonal_tree, dendro_tree, networkx_tree, newick_tree

__all__ = (heatmap, clonal_tree, dendro_tree, networkx_tree, newick_tree)



"""Solver Module."""

from collections import OrderedDict

import click

from scphylo.commands.solver._bnb import bnb
from scphylo.commands.solver._booster import booster
from scphylo.commands.solver._gpps import gpps
from scphylo.commands.solver._grmt import grmt
from scphylo.commands.solver._huntress import huntress
from scphylo.commands.solver._onconem import onconem
from scphylo.commands.solver._phiscs import phiscsb, phiscsi
from scphylo.commands.solver._scistree import scistree
from scphylo.commands.solver._scite import scite
from scphylo.commands.solver._siclonefit import siclonefit
from scphylo.commands.solver._sphyr import sphyr


class NaturalOrderGroup(click.Group):
    """Command group trying to list subcommands in the order they were added.

    Make sure you initialize the `self.commands` with OrderedDict instance.
    With decorator, use::
        @click.group(cls=NaturalOrderGroup, commands=OrderedDict())
    """

    def list_commands(self, ctx):
        """List command names as they are in commands dict.

        If the dict is OrderedDict, it will preserve the order commands
        were added.
        """
        return self.commands.keys()


@click.group(
    name="solver",
    cls=NaturalOrderGroup,
    commands=OrderedDict(),
    context_settings={"max_content_width": 300, "terminal_width": 300},
)
def cli_solver():
    """Solver module."""
    return None


cli_solver.add_command(booster)
cli_solver.add_command(phiscsb)
cli_solver.add_command(phiscsi)
cli_solver.add_command(scite)
cli_solver.add_command(scistree)
cli_solver.add_command(bnb)
cli_solver.add_command(huntress)
cli_solver.add_command(grmt)
cli_solver.add_command(sphyr)
cli_solver.add_command(gpps)
cli_solver.add_command(onconem)
cli_solver.add_command(siclonefit)



import glob
import os
import time

import numpy as np
import pandas as pd

import scphylo as scp


def siclonefit(
    df_input,
    alpha,
    beta,
    n_restarts=3,
    n_iters=500,
    n_burnin=100,
    return_tree=False,
    experiment=False,
):
    """Solving using SiCloneFit.

    Bayesian inference of population structure, genotype, and phylogeny of tumor clones
    from single-cell genome sequencing data :cite:`SiCloneFit`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing
        entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    n_restarts : :obj:`int`, optional
        Number of restarts, by default 3
    n_iters : :obj:`int`, optional
        Number of iterations for each Markov Chain after burnin, by default 500
    n_burnin : :obj:`int`, optional
        Number of iterations for burnin of each Markov Chain, by default 100
    return_tree : :obj:`bool`, optional
        Return the inferred cell-lineage tree, by default False
    experiment : :obj:`bool`, optional
        Is in the experiment mode (the log won't be shown), by default False

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    executable = scp.ul.executable("SiCloneFiTComplete.jar", "SiCloneFit")

    if not experiment:
        scp.logg.info(
            f"running SiCloneFit with alpha={alpha}, beta={beta}, n_iters={n_iters}"
        )

    tmpdir = scp.ul.tmpdirsys(suffix=".siclonefit")

    df_input.T.reset_index(drop=True).to_csv(
        f"{tmpdir.name}/siclonefit.input", sep=" ", header=None
    )
    with open(f"{tmpdir.name}/siclonefit.cellnames", "w") as fout:
        fout.write(" ".join(df_input.index))
    with open(f"{tmpdir.name}/siclonefit.genenames", "w") as fout:
        fout.write(" ".join(df_input.columns))
    I_mtr = df_input.values

    cmd = (
        f"java -jar {executable} "
        f"-m {df_input.shape[0]} "
        f"-n {df_input.shape[1]} "
        f"-ipMat {tmpdir.name}/siclonefit.input "
        f"-fp {alpha} "
        f"-fn {beta} "
        "-df 0 "
        f"-missing {np.sum(I_mtr == 3)/(I_mtr.size)} "
        f"-iter {n_iters} "
        f"-cellNames {tmpdir.name}/siclonefit.cellnames "
        f"-geneNames {tmpdir.name}/siclonefit.genenames "
        f"-r {n_restarts} "
        f"-burnin {n_burnin} "
        # "-recurProb 0 "
        # "-delProb 0 "
        # "-LOHProb 0 "
        # "-doublet "
        # "-printIter "
        # "-treeIter "
        f"-outDir {tmpdir.name} > {tmpdir.name}/siclonefit.log"
    )
    s_time = time.time()
    os.system(cmd)
    e_time = time.time()
    running_time = e_time - s_time

    out_dir = glob.glob(f"{tmpdir.name}/*samples/best")[0]

    df_output = pd.read_csv(
        f"{out_dir}/best_MAP_predicted_genotype.txt",
        sep=" ",
        header=None,
        index_col=0,
    ).T
    df_output.columns = df_input.columns
    df_output.index = df_input.index
    df_output.index.name = "cellIDxmutID"

    with open(f"{out_dir}/best_MAP_tree.txt") as fin:
        tree = fin.readline().strip()

    tmpdir.cleanup()

    if not experiment:
        scp.ul.stat(df_input, df_output, alpha, beta, running_time)
        if return_tree:
            return df_output, tree
        else:
            return df_output
    else:
        is_cf = scp.ul.is_conflict_free_gusfield(df_output)
        nll = scp.ul.calc_nll_matrix(df_input, df_output, alpha, beta)
        return df_output, running_time, is_cf, nll



"""Datasets Module."""

from scphylo.datasets._datasets import (
    acute_lymphocytic_leukemia1,
    acute_lymphocytic_leukemia2,
    acute_lymphocytic_leukemia3,
    acute_lymphocytic_leukemia4,
    acute_lymphocytic_leukemia5,
    acute_lymphocytic_leukemia6,
    acute_myeloid_leukemia1,
    acute_myeloid_leukemia2,
    colorectal1,
    colorectal2,
    colorectal3,
    erbc,
    example,
    high_grade_serous_ovarian_cancer1,
    high_grade_serous_ovarian_cancer2,
    high_grade_serous_ovarian_cancer3,
    high_grade_serous_ovarian_cancer_3celllines,
    melanoma20,
    muscle_invasive_bladder,
    myeloproliferative_neoplasms18,
    myeloproliferative_neoplasms78,
    myeloproliferative_neoplasms712,
    oligodendroglioma_idh_mutated_tumor,
    renal_cell_carcinoma,
    test,
    tnbc,
)
from scphylo.datasets._simulate import add_doublets, add_noise, add_readcount, simulate

__all__ = (
    acute_lymphocytic_leukemia1,
    acute_lymphocytic_leukemia2,
    acute_lymphocytic_leukemia3,
    acute_lymphocytic_leukemia4,
    acute_lymphocytic_leukemia5,
    acute_lymphocytic_leukemia6,
    acute_myeloid_leukemia1,
    acute_myeloid_leukemia2,
    colorectal1,
    colorectal2,
    colorectal3,
    erbc,
    example,
    high_grade_serous_ovarian_cancer1,
    high_grade_serous_ovarian_cancer2,
    high_grade_serous_ovarian_cancer3,
    high_grade_serous_ovarian_cancer_3celllines,
    melanoma20,
    muscle_invasive_bladder,
    myeloproliferative_neoplasms18,
    myeloproliferative_neoplasms78,
    myeloproliferative_neoplasms712,
    oligodendroglioma_idh_mutated_tumor,
    renal_cell_carcinoma,
    test,
    tnbc,
    simulate,
    add_noise,
    add_doublets,
    add_readcount,
)



import subprocess

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run Sequenza.")
@click.argument(
    "outdir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "normal",
    required=True,
    type=str,
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.option(
    "--time",
    default=["1:00:00", "20:00:00"],
    multiple=True,
    show_default=True,
    help="Time List [indexing, calling].",
)
@click.option(
    "--mem",
    default=["20", "20"],
    multiple=True,
    show_default=True,
    help="Memory List [indexing, calling].",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def sequenza(outdir, normal, ref, time, mem, afterok):
    """Run Sequenza.

    scphylo caller sequenza /path/to/in/dir normal_name hg19|hg38|mm10

    BAM files (*.markdup_bqsr.bam) --> folder (.sequenza)
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    def step1(afterok):
        def get_command():
            cmds = ""
            cmds += cmd([f"mkdir -p {outdir}/_indexing/sequenza"])
            cmds += cmd([f"module load {scp.settings.tools['sequenza']}"])
            cmds += cmd(
                [
                    "sequenza-utils",
                    "gc_wiggle",
                    "-w 50",
                    f"--fasta {config['ref']}",
                    f"-o {outdir}/_indexing/sequenza/sequenza.wig.gz",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        cmds_tmp = pd.DataFrame()
        cmds_tmp["cmd"] = [get_command()]
        cmdmain = write_cmds_get_main(
            cmds_tmp,
            "sequenza-1of2",
            time[0],
            mem[0],
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step2(afterok):
        def get_command(sample):
            cmds = ""
            cmds += cmd([f"module load {scp.settings.tools['sequenza']}"])
            cmds += cmd([f"module load {scp.settings.tools['samtools']}"])
            cmds += cmd(
                [
                    "sequenza-utils",
                    "bam2seqz",
                    f"--normal {outdir}/{normal}.markdup_bqsr.bam",
                    f"--tumor {outdir}/{sample}.markdup_bqsr.bam",
                    f"--fasta {config['ref']}",
                    f"-gc {outdir}/_indexing/sequenza/sequenza.wig.gz",
                    f"--output {outdir}/{sample}.seqz.gz",
                    # f"--chromosome {' '.join(chroms)}",
                ]
            )
            cmds += cmd(
                [
                    "sequenza-utils",
                    "seqz_binning",
                    f"--seqz {outdir}/{sample}.seqz.gz",
                    "-w 50",
                    f"-o {outdir}/{sample}.50.seqz.gz",
                ]
            )
            cmds += cmd(
                [
                    "Rscript",
                    f"{scp.ul.get_file('scphylo.ul/sequenza.R')}",
                    outdir,
                    sample,
                ]
            )
            # cmds += cmd(
            #     [
            #         "rm -rf",
            #         f"{outdir}/{sample}.50.seqz.gz",
            #         f"{outdir}/{sample}.50.seqz.gz.tbi",
            #         f"{outdir}/{sample}.seqz.gz",
            #         f"{outdir}/{sample}.seqz.gz.tbi",
            #     ]
            # )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        df_cmds_tmp = scp.ul.get_samples_df(outdir, normal)
        df_cmds_tmp["cmd"] = df_cmds_tmp.apply(
            lambda x: get_command(x["sample"]), axis=1
        )
        cmdmain = write_cmds_get_main(
            df_cmds_tmp,
            "sequenza-2of2",
            time[1],
            mem[1],
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    cmd1 = step1(afterok)
    code1 = subprocess.getoutput(cmd1)
    cmd2 = step2(code1)
    code2 = subprocess.getoutput(cmd2)
    scp.logg.info(code2)

    return None



import scphylo as scp


def binarym_filter_private_mutations(df):
    df.drop(df.columns[df.sum() == 1], axis=1, inplace=True)


def binarym_filter_clonal_mutations(df):
    x = (df == 1).sum()
    x = x[x == df.shape[0]]
    df.drop(x.index, axis=1, inplace=True)


def binarym_filter_nonsense_mutations(df, alt_in=2, ref_in=1):
    df.drop(
        df.columns[
            ~(
                ((df == 1).sum() >= alt_in)
                & (
                    ((df == 0).sum() >= ref_in)
                    | ((df == 1).sum() == df.shape[0])
                    | ((df == 1).sum() >= (df == 3).sum())
                )
            )
        ],
        axis=1,
        inplace=True,
    )


def binarym_statistics(df):
    t = df.shape[0] * df.shape[1]
    a = (df == 0).sum().sum()
    b = (df == 1).sum().sum()
    d = (df == 3).sum().sum()
    scp.logg.info(f"size = {df.shape[0]} × {df.shape[1]}")
    scp.logg.info(f"    REF     = {a:6d} ({100*a/t:2.1f}%)")
    scp.logg.info(f"    HET     = {b:6d} ({100*b/t:2.1f}%)")
    scp.logg.info(f"    UNKNOWN = {d:6d} ({100*d/t:2.1f}%)")


def consensus_combine(df):
    """Combine cells in genotype matrix.

    This function combines the replicates or cells that have
    the same origin prior to running Trisicell-Cons. The replicates
    or cells that are supposed to be merged must be designated
    with `_`. For instance:

    input: {`{Cell1}_{ID1}`, `{Cell1}_{ID2}`, `{Cell2}_{ID1}`, `{Cell2}_{ID2}`}.

    output: {`{Cell1}`, `{Cell2}`}.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        The input genotype matrix in conflict-free format.

    Returns
    -------
    :class:`pandas.DataFrame`
        The combine genotype matrix.
    """
    df2 = df.groupby(df.index.str.split("_").str[0]).transform("prod")
    df2 = df2.groupby(df2.index.str.split("_").str[0]).first()
    return df2



import itertools
import math

import scphylo as scp
from scphylo.external.gpps._utils_ilp import expand_name

__author__ = "Simone Ciccolella"
__date__ = "11/30/21"


def gpps_ilp(
    input_matrix,
    alpha,
    beta,
    k_dollo,
    max_del=-1,
    time_limit=86400,
    n_threads=1,
):
    gp, gp_is_not_imported = scp.ul.import_gurobi()
    if gp_is_not_imported:
        scp.logg.error("Unable to import a package!")

    # ==================================================================#
    # ========================= PREREQUISITES ==========================#
    # ==================================================================#

    # parser.add_argument(
    #     "--mps",
    #     action="store_true",
    #     help="This will output the model in MPS format instead of running the solver",
    # )

    # ----------------------Initialize program----------------------#
    # Fixed parameters
    # num_samples = len(input_matrix)
    # num_clones = int(num_mutations * n_clones)
    num_clones = len(input_matrix)
    num_mutations = len(input_matrix[0])

    # scp.logg.debug("Num samples: %d" % num_samples)
    scp.logg.debug("Num mutations: %d" % num_mutations)
    scp.logg.debug("Num clones: %d" % num_clones)

    # ==================================================================#
    # ========================== GUROBI MODEL ==========================#
    # ==================================================================#
    model = gp.Model("Parsimony Phylogeny Model")
    model.Params.OutputFlag = 0
    model.Params.LogFile = ""
    model.setParam("Threads", n_threads)
    model.setParam("TimeLimit", time_limit)

    # ---------------------------------------------------#
    # ------------------- VARIABLES ---------------------#
    # ---------------------------------------------------#

    # -----------Variable Y and B---------------

    lalpha = math.log(alpha)
    lbeta = math.log(beta)
    l_alpha = math.log(1 - alpha)
    l_beta = math.log(1 - beta)

    scp.logg.debug(lalpha, lbeta, l_alpha, l_beta)

    scp.logg.debug("Generating variables I, F, w, P.")
    I_mtx = {}
    F = {}
    fminus = {}
    P = {}
    # False positive should be only a few
    FP = {}

    for row_index, row in enumerate(input_matrix):
        I_mtx[row_index] = {}
        F[row_index] = {}
        fminus[row_index] = {}
        P[row_index] = {}
        FP[row_index] = {}
        for col_index, cell in enumerate(row):
            # P
            names = expand_name(str(col_index), 1, k_dollo)
            for name in names:
                P[row_index][name] = model.addVar(
                    vtype=gp.GRB.BINARY, obj=0, name=f"P{row_index}-{name}"
                )
            if cell < 2:
                # I
                I_mtx[row_index][col_index] = model.addVar(
                    vtype=gp.GRB.BINARY,
                    obj=0,
                    name=f"I{row_index}-{col_index}",
                )
                model.update()
                model.addConstr(
                    I_mtx[row_index][col_index] == cell,
                    f"constr_I{row_index}-{col_index}",
                )
                # F and fminus. fminus is equal to 1-F
                if cell == 0:
                    F[row_index][col_index] = model.addVar(
                        vtype=gp.GRB.BINARY,
                        obj=lalpha,
                        name=f"F{row_index}-{col_index}",
                    )
                    fminus[row_index][col_index] = model.addVar(
                        vtype=gp.GRB.BINARY,
                        obj=l_beta,
                        name=f"f{row_index}-{col_index}",
                    )
                if cell == 1:
                    F[row_index][col_index] = model.addVar(
                        vtype=gp.GRB.BINARY,
                        obj=l_alpha,
                        name=f"F{row_index}-{col_index}",
                    )
                    fminus[row_index][col_index] = model.addVar(
                        vtype=gp.GRB.BINARY,
                        obj=lbeta,
                        name=f"f{row_index}-{col_index}",
                    )
                    FP[row_index][col_index] = model.addVar(
                        vtype=gp.GRB.BINARY,
                        obj=0,
                        name=f"FP{row_index}-{col_index}",
                    )
                    model.update()
                    model.addConstr(
                        FP[row_index][col_index] == 1 - P[row_index][names[0]],
                        name=f"constr_FP{row_index}-{col_index}",
                    )
                model.update()
                model.addConstr(
                    F[row_index][col_index] == 1 - fminus[row_index][col_index],
                    name=f"constr_def_fminus{row_index}-{col_index}",
                )
                model.addConstr(
                    F[row_index][col_index]
                    == P[row_index][names[0]]
                    - gp.quicksum(P[row_index][name] for name in names[1:]),
                    name=f"constr_balance_F{row_index}-{col_index}",
                )
                model.addConstr(
                    P[row_index][names[0]]
                    >= gp.quicksum(P[row_index][name] for name in names[1:]),
                    name=f"constr_imbalance_P{row_index}-{col_index}",
                )

    # There are only a few false positives
    model.addConstr(
        4 >= gp.quicksum(FP[r][c] for r in FP.keys() for c in FP[r].keys()),
        name="constr_few_FP",
    )

    model.update()

    scp.logg.debug("Generating variables B.")
    B = {}
    columns = list(
        itertools.chain.from_iterable(
            [expand_name(str(name), 1, k_dollo) for name in range(num_mutations)]
        )
    )
    for p in columns:
        B[p] = {}
    for p, q in itertools.combinations(columns, 2):
        B[p][q] = {}
        B[p][q]["01"] = model.addVar(vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},0,1]")
        B[p][q]["10"] = model.addVar(vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},1,0]")
        B[p][q]["11"] = model.addVar(vtype=gp.GRB.BINARY, obj=0, name=f"B[{p},{q},1,1]")
        model.update()
        for row_index, _ in enumerate(input_matrix):
            model.addConstr(
                B[p][q]["01"] >= P[row_index][q] - P[row_index][p],
                f"constr_B01-{p}-{q}-{row_index}",
            )
            model.addConstr(
                B[p][q]["10"] >= P[row_index][p] - P[row_index][q],
                f"constr_B10-{p}-{q}-{row_index}",
            )
            model.addConstr(
                B[p][q]["11"] >= P[row_index][p] + P[row_index][q] - 1,
                f"constr_B11-{p}-{q}-{row_index}",
            )
        model.addConstr(
            B[p][q]["01"] + B[p][q]["10"] + B[p][q]["11"] <= 2,
            f"constr_sum_B{p}-{q}",
        )

    deletions = {}
    del_names = []
    for p in columns:
        if "-" in p:
            del_names.append(p)
            deletions[p] = model.addVar(vtype=gp.GRB.BINARY, obj=0, name=f"Del[{p}]")
            for row_index, _ in enumerate(input_matrix):
                model.addConstr(deletions[p] >= P[row_index][p])

    if max_del != -1:
        model.addConstr(gp.quicksum(deletions[x] for x in del_names) <= max_del)

    model.update()
    model.modelSense = gp.GRB.MAXIMIZE
    model.update()

    # ---------------------------------------------------#
    # -------------------- OPTIMIZE ---------------------#
    # ---------------------------------------------------#
    scp.logg.debug("#----- GUROBI OPTIMIZATION ----#")
    model.optimize()

    # ==================================================================#
    # ======================= POST OPTIMIZATION ========================#
    # ==================================================================#

    output_matrix = []
    for row_index, row in enumerate(input_matrix):
        row_out = []
        for col_index, _ in enumerate(row):
            names = expand_name(str(col_index), 1, k_dollo)
            for name in names:
                row_out.append(int(float(P[row_index][name].X)))
        output_matrix.append(row_out)

    return output_matrix



import os

import click

import scphylo as scp


@click.command(short_help="Run gpps.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
@click.option(
    "--k_dollo",
    "-k",
    default=0,
    type=int,
    show_default=True,
    help="k-Dollo.",
)
@click.option(
    "--max_del",
    "-d",
    default=-1,
    type=int,
    show_default=True,
    help="Maximum number of deletion allowed.",
)
@click.option(
    "--neighbor_size",
    "-s",
    default=30,
    type=int,
    show_default=True,
    help="Hill climbing neighborhood size.",
)
@click.option(
    "--n_iters",
    "-l",
    default=100,
    type=int,
    show_default=True,
    help="Hill climbing maximum iterations.",
)
@click.option(
    "--time_limit",
    "-t",
    default=86400,
    type=int,
    show_default=True,
    help="Time limit of the program (in second).",
)
@click.option(
    "--n_threads",
    "-p",
    default=1,
    type=int,
    show_default=True,
    help="Number of threads.",
)
def gpps(
    genotype_file,
    alpha,
    beta,
    k_dollo,
    max_del,
    neighbor_size,
    n_iters,
    time_limit,
    n_threads,
):
    """gpps.

    an ILP-based approach for inferring cancer progression with mutation losses from
    single cell data :cite:`gpps`.

    scphylo solver gpps input.SC 0.0001 0.1 -k 0 -s 30 -l 100 -t 86400 -p 1
    """
    outfile = os.path.splitext(genotype_file)[0]

    scp.settings.verbosity = "info"
    scp.settings.logfile = f"{outfile}.gpps.log"

    df_in = scp.io.read(genotype_file)
    if alpha == 0:
        alpha = 0.000000000001
    df_out = scp.tl.gpps(
        df_in,
        alpha=alpha,
        beta=beta,
        k_dollo=k_dollo,
        max_del=max_del,
        neighbor_size=neighbor_size,
        n_iters=n_iters,
        time_limit=time_limit,
        n_threads=n_threads,
    )
    scp.io.write(df_out, f"{outfile}.gpps.CFMatrix")

    return None
